"""
Create a map from one column of a stacked Jacobian to a full (dense or
sparse) Jacobian matrix
"""


#[
from __future__ import annotations

import itertools as it_

from typing import (TYPE_CHECKING, )
if TYPE_CHECKING:
    from typing import (Any, Self, )
    from collections.abc import (Iterable, Collection, )
#]


class ArrayMap:
    """
    """
    #[

    __slots__ = ('lhs', 'rhs', )

    def __init__(
        self,
        lhs: tuple[list[int], list[int]] | None = None,
        rhs: tuple[list[int], list[int]] | None = None,
    ) -> None:
        """
        """
        self.lhs = ([], [], ) if lhs is None else lhs
        self.rhs = ([], [], ) if rhs is None else rhs

    @classmethod
    def static(
        klass,
        eids: list[int],
        eid_to_wrt_tokens: dict[int, Any],
        tokens_in_columns_on_lhs: list[Any],
        eid_to_rhs_offset: dict[int, int],
        /,
        **kwargs,
    ) -> Self:
        """
        """
        self = klass()
        token_to_lhs_column = {
            t: i
            for i, t in enumerate(tokens_in_columns_on_lhs, )
        }
        for lhs_row, eid in enumerate(eids):
            self.add_lhs_rhs(*_get_raw_map_for_single_equation(
                eid_to_wrt_tokens[eid],
                token_to_lhs_column,
                eid_to_rhs_offset[eid],
                lhs_row,
                **kwargs,
            ))
        return self

    def __len__(self, /, ) -> int:
        return len(self.lhs[0])

    def append(
        self,
        lhs: tuple[list[int], list[int]],
        rhs: tuple[list[int], list[int]],
        /
    ) -> None:
        """
        """
        self.lhs[0].append(lhs[0])
        self.lhs[1].append(lhs[1])
        self.rhs[0].append(rhs[0])
        self.rhs[1].append(rhs[1])

    def add_lhs_rhs(
        self,
        lhs_rows: Iterable[int],
        lhs_columns: Iterable[int],
        rhs_rows: Iterable[int],
        rhs_columns: Iterable[int],
        /,
    ) -> None:
        """
        """
        self.lhs = (self.lhs[0]+list(lhs_rows), self.lhs[1]+list(lhs_columns), )
        self.rhs = (self.rhs[0]+list(rhs_rows), self.rhs[1]+list(rhs_columns), )

    def remove_nones(self) -> None:
        """
        Remove any map entry that has a None for the row index on the LHS
        """
        if not self.lhs[0]:
            return
        zipped_pruned = [
            i for i in zip(self.lhs[0], self.lhs[1], self.rhs[0], self.rhs[1])
            if i[0] is not None
        ]
        unzipped_pruned = list(zip(*zipped_pruned))
        self.lhs = (list(unzipped_pruned[0]), list(unzipped_pruned[1]))
        self.rhs = (list(unzipped_pruned[2]), list(unzipped_pruned[3]))


class VectorMap:
    """
    """
    #[

    __slots__ = ('lhs', 'rhs', )

    def __init__(self, /, ) -> None:
        """
        """
        self.lhs = ([], )
        self.rhs = ([], )

    @classmethod
    def static(
        klass,
        eids: Collection[int],
    ) -> Self:
        """
        """
        self = klass()
        num_equations = len(eids)
        self.lhs = (list(range(num_equations)), )
        self.rhs = (list(eids), )
        return self


    def __len__(self, /, ) -> int:
        return len(self.lhs[0])

    def append(
        self,
        lhs: tuple[list[int]],
        rhs: tuple[list[int]],
        /
    ) -> None:
        """
        """
        self.lhs[0].append(lhs[0])
        self.rhs[0].append(rhs[0])

    #]


def create_eid_to_rhs_offset(
    eids: tuple[int],
    eid_to_wrt_something: dict[int, Any],
    /,
) -> dict[int, int]:
    """
    Cumulative sum of number of tokens in individual equations, starting with 0
    """
    rhs_offsets = list(it_.accumulate(
        len(eid_to_wrt_something[i]) 
        for i in eids
    ))
    # Offset starts from 0 for the first equations
    rhs_offsets.pop()
    rhs_offsets.insert(0, 0)
    return dict(zip(eids, rhs_offsets))


def _get_raw_map_for_single_equation(
    tokens_in_equation_on_rhs: in_.Tokens,
    token_to_lhs_column: dict[Any, int],
    rhs_offset: int,
    lhs_row: int,
    /,
    rhs_column: int,
    lhs_column_offset: int,
    **kwargs,
) -> tuple[tuple[list[int], list[int]], tuple[list[int], list[int]]]:
    """
    """
    raw_map = (
        (lhs_row, lhs_column_offset + token_to_lhs_column[t], rhs_row, rhs_column, )
        for rhs_row, t in enumerate(tokens_in_equation_on_rhs, start=rhs_offset)
        if t in token_to_lhs_column.keys()
    )
    # Collect all lhr_rows, all lhs_columns, all rhs_rows, all rhs_columns
    raw_map = zip(*raw_map)
    try:
        lhs_rows = next(raw_map, )
        lhs_columns = next(raw_map, )
        rhs_rows = next(raw_map, )
        rhs_columns = next(raw_map, )
    except:
        lhs_rows = ()
        lhs_columns = ()
        rhs_rows = ()
        rhs_columns = ()
    return lhs_rows, lhs_columns, rhs_rows, rhs_columns




"""
Model source
"""


#[

from __future__ import annotations

import re as _re
from typing import Self, Any, Type, TypeAlias, Literal, Protocol, NoReturn
from collections.abc import Iterable
import documark as _dm

from . import equations as _equations
from .equations import Equation, EquationKind
from . import quantities as _quantities
from .quantities import QuantityKind, Quantity
from . import wrongdoings as _wd
from .parsers import preparser as _preparser
from .parsers import models as _models
from .parsers import common as _common

#]


__all__ = (
    "ModelSource",
)


Description: TypeAlias = str
Human: TypeAlias = str
Attributes: TypeAlias = tuple[str, ...]
QuantityInput: TypeAlias = tuple[Description, Human, Attributes]
EquationInput: TypeAlias = tuple[Description, tuple[Human, Human], Attributes]


LOGGABLE_VARIABLE = (
    QuantityKind.TRANSITION_VARIABLE
    | QuantityKind.MEASUREMENT_VARIABLE
    | QuantityKind.EXOGENOUS_VARIABLE
)


LOGGABLE_VARIABLE_OR_ANY_SHOCK = (
    LOGGABLE_VARIABLE
    | QuantityKind.ANY_SHOCK
)


class SourceMixinProtocol(Protocol, ):
    """
    """
    #[

    @classmethod
    def from_source(
        klass,
        source: ModelSource,
        /,
        **kwargs,
    ) -> Self:
        ...

    #]


def from_string(
    klass: type[SourceMixinProtocol],
    source_string: str,
    /,
    context: dict | None = None,
    save_preparsed: str = "",
    **kwargs,
) -> SourceMixinProtocol:
    """Create a new object from model source string"""
    #[
    source, info = ModelSource.from_string(
        source_string, context=context, save_preparsed=save_preparsed,
    )
    return klass.from_source(source, context=context, **kwargs, )
    #]


def from_file(
    klass: type[SourceMixinProtocol],
    source_files: str | Iterable[str],
    /,
    **kwargs,
) -> SourceMixinProtocol:
    """Create a new object from source file(s)"""
    #[
    source_string = _combine_source_files_into_string(source_files, )
    return from_string(klass, source_string, **kwargs, )
    #]


class ModelSource:
    """
    """
    #[

    __slots__ = (
        "quantities",
        "dynamic_equations",
        "steady_equations",
        "context",
    )

    def __init__(self, /, ) -> None:
        self.quantities = []
        self.dynamic_equations = []
        self.steady_equations = []
        self.context = None

    @property
    def num_quantities(self, /, ) -> int:
        return len(self.quantities)

    @property
    def all_names(self, /, ) -> Iterable[str]:
        return [ qty.human for qty in self.quantities ]

    def _add_parameters(self, names: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(names, QuantityKind.PARAMETER)

    def _add_exogenous_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, QuantityKind.EXOGENOUS_VARIABLE)

    def _add_transition_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, QuantityKind.TRANSITION_VARIABLE)

    def _add_anticipated_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, QuantityKind.ANTICIPATED_SHOCK)

    def _add_unanticipated_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, QuantityKind.UNANTICIPATED_SHOCK)

    def _add_transition_equations(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, EquationKind.TRANSITION_EQUATION)

    def _add_measurement_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, QuantityKind.MEASUREMENT_VARIABLE)

    def _add_measurement_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, QuantityKind.MEASUREMENT_SHOCK)

    def _add_measurement_equations(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, EquationKind.MEASUREMENT_EQUATION)

    def _add_steady_autovalues(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, EquationKind.STEADY_AUTOVALUES, )

    def _add_all_but(self, all_but_input: str | None) -> None:
        if not all_but_input:
            return
        self.all_but.append(all_but_input)

    def _add_log_variables(self, log_variable_inputs: Iterable[str] | None, /, ) -> None:
        if not log_variable_inputs:
            return
        self.log_variables += log_variable_inputs

    def _add_quantities(self, quantity_inputs: Iterable[QuantityInput] | None, kind: QuantityKind) -> None:
        """
        """
        if not quantity_inputs:
            return
        start_entry = self.quantities[-1].entry + 1 if self.quantities else 0
        self.quantities = self.quantities + [
            Quantity(
                id=None,
                human=q[1].strip(),
                kind=kind,
                logly=None,
                description=q[0].strip(),
                entry=i,
                attributes=_conform_attributes(q[2], )
            )
            for i, q in enumerate(quantity_inputs, start=start_entry, )
        ]

    def _add_equations(self, equation_inputs: Iterable[EquationInput] | None, kind: EquationKind, /, ) -> None:
        """
        """
        if not equation_inputs:
            return
        #
        def _create_equations(equation_inputs, kind, start_entry, human_func):
            return [
                Equation(
                    id=None,
                    human=human_func(ein),
                    kind=kind,
                    description=ein[0].strip(),
                    entry=i,
                    attributes=_conform_attributes(ein[2], ),
                )
                for i, ein in enumerate(equation_inputs, start=start_entry)
            ]
        #
        def _human_func_dynamic(ein):
            return _handle_white_spaces(ein[1][0])
        #
        def _human_func_steady(ein):
            return _handle_white_spaces(ein[1][1] if ein[1][1] else ein[1][0])
        #
        equation_inputs = tuple(equation_inputs)
        start_entry = self.dynamic_equations[-1].entry + 1 if self.dynamic_equations else 0
        self.dynamic_equations += _create_equations(equation_inputs, kind, start_entry, _human_func_dynamic)
        self.steady_equations += _create_equations(equation_inputs, kind, start_entry, _human_func_steady)

    def _verify_log_variables(self, log_variables: Iterable[str] | None) -> None:
        """
        """
        if not log_variables:
            return
        loggables = _extract_loggable_names(self.quantities, )
        illegal_log_variables = set(log_variables) - loggables
        if illegal_log_variables:
            raise _wd.IrisPieCritical((
                f"Illegal name(s) on the {_models.LOG_VARIABLES_KEYWORD} list",
                *illegal_log_variables,
            ))

    def _populate_logly(
        self,
        log_variables: Iterable[str] | None,
        all_but: bool | False,
    ) -> None:
        """
        """
        log_variables = set(log_variables) if log_variables else set()
        unlisted_logly = bool(all_but)
        listed_logly = not unlisted_logly
        def is_logly(human: str, /, ) -> bool:
            return listed_logly if human in log_variables else unlisted_logly
        for qty in self.quantities:
            if qty.kind not in LOGGABLE_VARIABLE:
                continue
            qty.logly = is_logly(qty.human, )

    @classmethod
    def from_lists(
        klass,
        transition_variables: Iterable[QuantityInput],
        transition_equations: Iterable[EquationInput],
        anticipated_shocks: Iterable[QuantityInput] | None = None,
        unanticipated_shocks: Iterable[QuantityInput] | None = None,
        measurement_variables: Iterable[QuantityInput] | None = None,
        measurement_equations: Iterable[EquationInput] | None = None,
        measurement_shocks: Iterable[QuantityInput] | None = None,
        steady_autovalues: Iterable[EquationInput] | None = None,
        parameters: Iterable[QuantityInput] | None = None,
        exogenous_variables: Iterable[QuantityInput] | None = None,
        log_variables: Iterable[str] | None = None,
        all_but: bool = False,
        context: dict | None = None,
    ) -> Self:
        """
        """
        self = klass()
        #
        self._add_transition_variables(transition_variables, )
        self._add_transition_equations(transition_equations, )
        self._add_anticipated_shocks(anticipated_shocks, )
        self._add_unanticipated_shocks(unanticipated_shocks, )
        self._add_measurement_variables(measurement_variables, )
        self._add_parameters(parameters, )
        self._add_exogenous_variables(exogenous_variables, )
        self._verify_log_variables(log_variables, )
        self._populate_logly(log_variables, all_but, )
        #
        self._add_measurement_equations(measurement_equations, )
        self._add_measurement_shocks(measurement_shocks, )
        self._add_steady_autovalues(steady_autovalues, )
        #
        self.context = (dict(context) if context else {}) | {"__builtins__": {}}
        #
        return self

    @classmethod
    def from_string(
        klass,
        source_string: str,
        *,
        context: dict | None = None,
        save_preparsed: str = "",
    ) -> tuple[Self, dict[str, Any]]:
        """
        """
        context = (dict(context) if context else {}) | {"__builtins__": {}}
        preparsed_string, preparser_info = _preparser.from_string(
            source_string, context=context, save_preparsed=save_preparsed, 
        )
        parsed_content = _models.from_string(preparsed_string)
        all_but = _is_all_but_present(parsed_content.get("all-but", ), )
        log_variables = set(parsed_content.get("log-variables", ()), )
        #
        self = klass.from_lists(
            transition_variables=parsed_content.get("transition-variables", ),
            transition_equations=parsed_content.get("transition-equations", ),
            anticipated_shocks=parsed_content.get("anticipated-shocks", ),
            unanticipated_shocks=parsed_content.get("unanticipated-shocks", ),
            measurement_variables=parsed_content.get("measurement-variables", ),
            measurement_equations=parsed_content.get("measurement-equations", ),
            measurement_shocks=parsed_content.get("measurement-shocks", ),
            steady_autovalues=parsed_content.get("steady-autovalues", ),
            parameters=parsed_content.get("parameters", ),
            exogenous_variables=parsed_content.get("exogenous-variables", ),
            log_variables=log_variables,
            all_but=all_but,
            context=context,
        )
        #
        return self, preparser_info

    @classmethod
    def from_file(
        klass,
        source_files: str | Iterable[str],
        /,
        **kwargs,
    ) -> Self:
        """Create a new ModelSource from source file(s)"""
        source_string = _combine_source_files_into_string(source_files, )
        return klass.from_string(source_string, **kwargs, )

    #]


def _handle_white_spaces(x: str, /, ) -> str:
    return _re.sub(r"[\s\n\r]+", "", x)


def _conform_attributes(attributes: str | Iterable[str] | None, /, ) -> set:
    """
    """
    #[
    if not attributes:
        return set()
    if isinstance(attributes, str):
        attributes = set((attributes, ))
    return set(a.strip() for a in attributes)
    #]


def _combine_source_files_into_string(
    source_files: str | Iterable[str],
    /,
    joint: str = "\n\n",
) -> str:
    """
    """
    #[
    if isinstance(source_files, str):
        source_files = [source_files]
    source_strings = []
    for f in source_files:
        with open(f, "r") as fid:
            source_strings.append(fid.read(), )
    return "\n\n".join(source_strings, )
    #]


def _verify_all_but(all_but: list[str] | None, /, ) -> None | NoReturn:
    if all_but is None:
        return
    if all(i == _models.ALL_BUT_KEYWORD for i in all_but):
        return
    if all(i == "" for i in all_but):
        return
    raise Exception("Inconsistent use of !all-but in !log-variables")


def _is_all_but_present(all_but: Iterable[str] | None, /, ) -> bool:
    if not all_but:
        return False
    all_but = tuple(all_but)
    _verify_all_but(all_but, )
    return all_but[0] == _models.ALL_BUT_KEYWORD


def _extract_loggable_names(quantities: Iterable[Quantity], /, ) -> set[str]:
    """
    """
    return set(
        i.human for i in quantities
        if i.kind in LOGGABLE_VARIABLE
    )


def serialize_to_portable(source_like, /, ) -> dict[str, Any]:
    """
    """
    #[
    portable_quantities = _serialize_quantities_to_portable(source_like.quantities, )
    portable_equations = _serialize_equations_to_portable(source_like.dynamic_equations, source_like.steady_equations, )
    return portable_quantities | portable_equations
    #]


_SERIALIZABLE_QUANTITY_KINDS = (
    QuantityKind.TRANSITION_VARIABLE,
    QuantityKind.MEASUREMENT_VARIABLE,
    QuantityKind.UNANTICIPATED_SHOCK,
    QuantityKind.ANTICIPATED_SHOCK,
    QuantityKind.MEASUREMENT_SHOCK,
    QuantityKind.PARAMETER,
    QuantityKind.EXOGENOUS_VARIABLE,
)


def _serialize_quantities_to_portable(quantities: tuple[Quantity], ) -> tuple[Quantity]:
    """
    """
    #[
    portable_quantities = {
        kind.to_portable(): _get_portable_quantities_of_kind(quantities, kind, )
        for kind in _SERIALIZABLE_QUANTITY_KINDS
    }
    return { k: v for k, v in portable_quantities.items() if v }
    #]


def _get_portable_quantities_of_kind(
    quantities: tuple[Quantity],
    kind: QuantityKind,
    /,
) -> tuple[tuple, ...]:
    """
    """
    return tuple(
        _serialize_quantity_to_portable(qty, )
        for qty in _quantities.generate_quantities_of_kind(quantities, kind, )
    )


def _serialize_quantity_to_portable(quantity: Quantity, /, ) -> tuple[str, bool, str, str, ]:
    """
    """
    return quantity.human, quantity.logly, quantity.description, " ".join(quantity.attributes, )


_SERIALIZABLE_EQUATION_KINDS = (
    EquationKind.TRANSITION_EQUATION,
    EquationKind.MEASUREMENT_EQUATION,
    EquationKind.STEADY_AUTOVALUES,
)


def _serialize_equations_to_portable(
    dynamic_equations: tuple[Equation],
    steady_equations: tuple[Equation],
    /,
) -> tuple[tuple, ...]:
    """
    """
    #[
    zipped = tuple(zip(dynamic_equations, steady_equations, ))
    portable_equations = {
        kind.to_portable(): _get_portable_equations_of_kind(zipped, kind, )
        for kind in _SERIALIZABLE_EQUATION_KINDS
    }
    return { k: v for k, v in portable_equations.items() if v }
    #]


def _get_portable_equations_of_kind(
    zipped_equations: tuple[tuple[Equation, Equation], ...],
    kind: EquationKind,
    /,
) -> tuple[tuple, ...]:
    """
    """
    return tuple(
        _serialize_equation_to_portable(dynamic, steady, )
        for dynamic, steady in zipped_equations
        if dynamic.kind == kind
    )


def _serialize_equation_to_portable(
    dynamic: Equation,
    steady: Equation,
    /,
) -> tuple[str, str, str, str, ]:
    #[
    dynamic_human = dynamic.human
    steady_human = steady.human
    if dynamic_human == steady_human:
        steady_human = ""
    return dynamic_human, steady_human, dynamic.description, " ".join(dynamic.attributes, )
    #]





"""
# First-order unsolved system matrices

## Unsolved system

$$
A E[x_t] + B E[x_{t-1}] + C + D u_t + E v_t = 0 \\
$$

$$
F y_t + G x_t + H + J w_t = 0
$$
"""


#[
from __future__ import annotations

import dataclasses as _dc
import numpy as _np

from . import descriptors as _descriptors
from ..simultaneous import _flags as _flags
#]


class System:
    """
    Unsolved system matrices
    """
    #[

    __slots__ = (
        # Transition equations
        "A", "B", "C", "D", "E",

        # Measurement equations
        "F", "G", "H", "J",
    )

    def __init__(
        self,
        descriptor: _descriptors.Descriptor,
        data_array: _np.ndarray,
        model_flags: _flags.ModelFlags,
        data_array_lagged: _np.ndarray | None,
        column_offset: int,
        /,
    ) -> None:
        """
        """
        #
        # Differentiate and evaluate constant
        #
        td, tc = descriptor.aldi_context.eval_to_arrays(data_array, column_offset, )

        smap = descriptor.system_map
        svec = descriptor.system_vectors

        self.A = _np.zeros(svec.shape_A_excl_dynid, dtype=float)
        self.A[smap.A.lhs] = td[smap.A.rhs]
        self.A = _np.vstack((self.A, smap.dynid_A))

        self.B = _np.zeros(svec.shape_B_excl_dynid, dtype=float)
        self.B[smap.B.lhs] = td[smap.B.rhs]
        self.B = _np.vstack((self.B, smap.dynid_B))

        if model_flags.is_linear:
            self.C = _np.zeros(svec.shape_C_excl_dynid, dtype=float)
            self.C[smap.C.lhs] = tc[smap.C.rhs, 0]
            self.C = _np.concatenate((self.C, smap.dynid_C))
        else:
            tokens = descriptor.system_vectors.transition_variables
            logly = descriptor.system_vectors.transition_variables_are_logly
            xi = _get_vector(descriptor, data_array, tokens, logly, column_offset, )
            xi_lagged = _get_vector(descriptor, data_array_lagged, tokens, logly, column_offset, )
            self.C = -(self.A @ xi + self.B @ xi_lagged)

        self.D = _np.zeros(svec.shape_D_excl_dynid, dtype=float)
        self.D[smap.D.lhs] = td[smap.D.rhs]
        self.D = _np.vstack((self.D, smap.dynid_D))

        self.E = _np.zeros(svec.shape_E_excl_dynid, dtype=float)
        self.E[smap.E.lhs] = td[smap.E.rhs]
        self.E = _np.vstack((self.E, smap.dynid_E))

        self.F = _np.zeros(svec.shape_F, dtype=float)
        self.F[smap.F.lhs] = td[smap.F.rhs]

        self.G = _np.zeros(svec.shape_G, dtype=float)
        self.G[smap.G.lhs] = td[smap.G.rhs]

        if model_flags.is_linear:
            self.H = _np.zeros(svec.shape_H, dtype=float)
            self.H[smap.H.lhs] = tc[smap.H.rhs, 0]
        else:
            tokens = descriptor.system_vectors.transition_variables
            logly = descriptor.system_vectors.transition_variables_are_logly
            xi = _get_vector(descriptor, data_array, tokens, logly, column_offset, )
            tokens = descriptor.system_vectors.measurement_variables
            logly = descriptor.system_vectors.measurement_variables_are_logly
            y = _get_vector(descriptor, data_array, tokens, logly, column_offset, )
            self.H = -(self.F @ y + self.G @ xi)

        self.J = _np.zeros(svec.shape_J, dtype=float)
        self.J[smap.J.lhs] = td[smap.J.rhs]


def _get_vector(
    descriptor: _descriptors.Descriptor,
    data_array: _np.ndarray,
    tokens: Iterable[_incidence.Token],
    logly: Iterable[bool],
    column_offset: int,
    /,
) -> _np.ndarray:
    """
    """
    #[
    logly = list(logly)
    rows = tuple(tok.qid for tok in tokens)
    columns = tuple(column_offset + tok.shift for tok in tokens)
    x = data_array[rows, columns]
    xx = _np.copy(x)
    x[logly] = _np.log(x[logly])
    return x
    #]




"""
"""


#[
from __future__ import annotations

from typing import (TYPE_CHECKING, )

from ..conveniences import descriptions as _descriptions
from ..quantities import (Quantity, QuantityKind, )

if TYPE_CHECKING:
    from collections.abc import (Iterable, )
#]


class Invariant(
    _descriptions.DescriptionMixin,
):
    """
    """
    #[

    __slots__ = (
        "quantities",
        "order",
        "has_constant",
    )

    def __init__(
        self,
        endogenous_names: Iterable[str],
        *,
        exogenous_names: Iterable[str] | None = None,
        order: int = 1,
        constant: bool = True,
    ) -> None:
        """
        """
        exogenous_names = exogenous_names or ()
        self.quantities = _create_quantities(endogenous_names, exogenous_names, )
        self.order = int(order)
        self.has_constant = bool(constant)

    #]


def _create_quantities(
    endogenous_names: Iterable[str],
    exogenous_names: Iterable[str] | None = None,
) -> tuple[Quantity, ...]:
    """
    """
    endogenous_names = tuple(endogenous_names)
    exogenous_names = tuple(exogenous_names or ())
    #
    endogenous_quantities = tuple(
        Quantity(id=qid, human=name, kind=QuantityKind.TRANSITION_VARIABLE, )
        for qid, name in enumerate(endogenous_names)
    )
    num_endogenous = len(endogenous_quantities)
    exogenous_quantities = tuple(
        Quantity(id=qid, human=name, kind=QuantityKind.EXOGENOUS_VARIABLE, )
        for qid, name in enumerate(exogenous_names, start=num_endogenous, )
    )
    return endogenous_quantities + exogenous_quantities




"""
Steady state evaluator
"""


#[

from __future__ import annotations

import itertools as _it
import numpy as _np

from ..evaluators.base import DEFAULT_INIT_GUESS
from .. import equations as _equations
from .. import quantities as _quantities

from . import _equators as _equators
from . import _jacobian as _jacobian
from . import _iter_printers as _iter_printers

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Iterable
    from numbers import Number
    from ..simultaneous._variants import Variant

#]


NONFLAT_STEADY_SHIFT = 1


class SteadyEvaluator:
    """
    """
    #[

    _equator_factory = ...

    _jacobian_factory = ...

    _iter_printer_factory = ...

    def __init__(
        self,
        wrt_qids_levels: Iterable[int],
        wrt_qids_changes: Iterable[int],
        wrt_equations: Iterable[_equations.Equation],
        all_quantities: Iterable[_quantities.Quantity],
        variant: Variant,
        *,
        context: dict | None = None,
        iter_printer_settings: dict[str, Any] | None = None,
    ) -> None:
        """
        """
        wrt_equations = list(wrt_equations, )
        #
        # Create an overall long list `self.wrt_qids` comprising qids from both
        # levels and changes, and then logical indices
        # `self._bool_index_wrt_levels` and `._bool_index_wrt_changes` for
        # accessing the levels and changes within the long list
        #
        self._merge_levels_and_changes(wrt_qids_levels, wrt_qids_changes, )
        #
        qid_to_logly = _quantities.create_qid_to_logly(all_quantities, )
        qid_to_name = _quantities.create_qid_to_name(all_quantities, )
        #
        # Reset changes to zero (only if flat)
        self._reset_changes(variant, qid_to_logly, )
        #
        shift_vec, self._min_shift = _prepare_time_shifts(wrt_equations, )
        self._shift_vec = shift_vec
        self._num_columns = shift_vec.shape[1]
        #
        # Index of loglies within wrt_qids; needs to be list not tuple
        # because of numpy indexing
        self._where_logly = \
            list(_quantities.generate_where_logly(self.wrt_qids, qid_to_logly, ))
        #
        # Vectors self._maybelog_init_levels and
        # self._maybelog_init_changes are the initial guesses from the
        # variant data with missing values filled in for the long list of
        # wrt_qids. The actual values iterated over are
        # self.maybelog_levels[self._bool_index_wrt_levels] and
        # self.maybelog_changes[self._bool_index_wrt_changes].
        #
        maybelog_levels, maybelog_changes = variant.retrieve_maybelog_values_for_qids(self.wrt_qids, qid_to_logly, )
        maybelog_levels, maybelog_changes = _fill_missing(maybelog_levels, maybelog_changes, )
        self._maybelog_init_levels = maybelog_levels
        self._maybelog_init_changes = maybelog_changes
        #
        self._num_levels = sum(self._bool_index_wrt_levels)
        self._num_changes = sum(self._bool_index_wrt_changes)
        #
        self._init_guess = _np.hstack((
            self._maybelog_init_levels[self._bool_index_wrt_levels],
            self._maybelog_init_changes[self._bool_index_wrt_changes],
        ))
        #
        self._steady_array = variant.create_steady_array(qid_to_logly, num_columns=shift_vec.shape[1], shift_in_first_column=self._min_shift, )
        self._update_steady_array(self._init_guess)
        #
        # Set up components
        self._equator = self._equator_factory(
            wrt_equations,
            context=context,
        )
        self._jacobian = self._jacobian_factory(
            wrt_equations,
            self.wrt_qids,
            qid_to_logly,
            context=context,
        )
        #
        self.iter_printer = self._iter_printer_factory(
            equations=wrt_equations,
            qids=self.wrt_qids,
            qid_to_logly=qid_to_logly,
            qid_to_name=qid_to_name,
            **(iter_printer_settings or {}),
        )
        #
        self.final_guess = None

    @property
    def _column_offset(self, /, ) -> int:
        """
        """
        return -self._min_shift

    def get_init_guess(self, *args, **kwargs, ) -> _np.ndarray:
        """
        """
        return self._init_guess

    def eval(
        self,
        maybelog_guess: _np.ndarray,
        /,
    ) -> tuple[_np.ndarray, _np.ndarray]:
        """
        """
        self._update_steady_array(maybelog_guess, )
        equator = self._equator.eval(self._steady_array, self._column_offset, )
        jacobian = self._jacobian.eval(self._steady_array, self._column_offset, )
        jacobian = jacobian[:, self._bool_index_wrt_levels + self._bool_index_wrt_changes]
        self.iter_printer.next(maybelog_guess, equator, jacobian_calculated=True, )
        return equator, jacobian

    def _merge_levels_and_changes(
        self,
        wrt_qids_levels: Iterable[str],
        wrt_qids_changes: Iterable[str],
        /,
    ) -> None:
        """
        """
        ...

    def _update_steady_array(
        self,
        current_guess: _np.ndarray,
        /,
    ) -> None:
        """
        """
        ...

    def extract_levels(
        self,
        guess: _np.ndarray,
        /,
    ) -> tuple[_np.ndarray, tuple[int, ...]]:
        """
        """
        levels = self._get_maybelog_levels(guess, )
        levels[self._where_logly] = _np.exp(levels[self._where_logly])
        levels = tuple(levels[self._bool_index_wrt_levels].tolist())
        wrt_qids_levels = tuple(_it.compress(self.wrt_qids, self._bool_index_wrt_levels, ))
        return levels, wrt_qids_levels,

    def extract_changes(
        self,
        guess: _np.ndarray,
        /,
    ) -> tuple[_np.ndarray, tuple[int, ...]]:
        """
        """
        changes = self._get_maybelog_changes(guess, )
        changes[self._where_logly] = _np.exp(changes[self._where_logly])
        changes = tuple(changes[self._bool_index_wrt_changes].tolist())
        wrt_qids_changes = tuple(_it.compress(self.wrt_qids, self._bool_index_wrt_changes, ))
        return changes, wrt_qids_changes,

    #]


class FlatSteadyEvaluator(SteadyEvaluator, ):
    """
    """
    #[

    _equator_factory = _equators.FlatSteadyEquator

    _jacobian_factory = _jacobian.FlatSteadyJacobian

    _iter_printer_factory = _iter_printers.FlatIterPrinter

    def _reset_changes(
        self,
        variant: Variant,
        qid_to_logly: dict[int, bool],
    ) -> None:
        """
        """
        variant.zero_changes(qid_to_logly, )

    def _merge_levels_and_changes(
        self,
        wrt_qids_levels: list[int],
        wrt_qids_changes: list[int],
        /,
    ) -> None:
        """
        """
        self.wrt_qids = list(set(wrt_qids_levels))
        self._bool_index_wrt_levels = [qid in wrt_qids_levels for qid in self.wrt_qids]
        self._bool_index_wrt_changes = []

    def _get_maybelog_levels(
        self,
        current_guess: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        new_maybelog_levels = _np.copy(self._maybelog_init_levels)
        if self._bool_index_wrt_levels:
            new_maybelog_levels[self._bool_index_wrt_levels] = current_guess
        return new_maybelog_levels

    def _get_maybelog_changes(
        self,
        current_guess: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        return _np.zeros(self._maybelog_init_changes.shape, dtype=float, )

    def _update_steady_array(
        self,
        current_guess: _np.ndarray,
        /,
    ) -> None:
        """
        """
        new_maybelog_levels = self._get_maybelog_levels(current_guess, )
        new_paths = _np.repeat(new_maybelog_levels.reshape(-1, 1), self._num_columns, axis=1, )
        new_paths[self._where_logly, :] = _np.exp(new_paths[self._where_logly, :])
        self._steady_array[self.wrt_qids, :] = new_paths
    #]


class NonflatSteadyEvaluator(SteadyEvaluator, ):
    """
    """
    #[

    _equator_factory = _equators.NonflatSteadyEquator
    _equator_factory.NONFLAT_STEADY_SHIFT = NONFLAT_STEADY_SHIFT

    _jacobian_factory = _jacobian.NonflatSteadyJacobian
    _jacobian_factory.NONFLAT_STEADY_SHIFT = NONFLAT_STEADY_SHIFT

    _iter_printer_factory = _iter_printers.NonflatIterPrinter

    def _reset_changes(self, *args, **kwargs, ) -> None:
        """
        """
        pass

    def _merge_levels_and_changes(
        self,
        wrt_qids_levels: list[str],
        wrt_qids_changes: list[str],
        /,
    ) -> None:
        self.wrt_qids = list(set(wrt_qids_levels) | set(wrt_qids_changes))
        self._bool_index_wrt_levels = [qid in wrt_qids_levels for qid in self.wrt_qids]
        self._bool_index_wrt_changes = [qid in wrt_qids_changes for qid in self.wrt_qids]

    def _get_maybelog_levels(
        self,
        current_guess: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        new_maybelog_levels = _np.copy(self._maybelog_init_levels)
        if self._bool_index_wrt_levels:
            new_maybelog_levels[self._bool_index_wrt_levels] = current_guess[:self._num_levels]
        return new_maybelog_levels

    def _get_maybelog_changes(
        self,
        current_guess: _np.ndarray,
        /,
    ) -> _np.ndarray:
        """
        """
        new_maybelog_changes = _np.copy(self._maybelog_init_changes)
        if self._bool_index_wrt_changes:
            new_maybelog_changes[self._bool_index_wrt_changes] = current_guess[self._num_levels:]
        return new_maybelog_changes

    def _update_steady_array(
        self,
        current_guess: _np.ndarray,
        /,
    ) -> None:
        """
        """
        new_maybelog_levels = self._get_maybelog_levels(current_guess, )
        new_maybelog_changes = self._get_maybelog_changes(current_guess, )
        new_paths = new_maybelog_levels.reshape(-1, 1) + self._shift_vec * new_maybelog_changes.reshape(-1, 1)
        new_paths[self._where_logly, :] = _np.exp(new_paths[self._where_logly, :])
        self._steady_array[self.wrt_qids, :] = new_paths
    #]


def _fill_missing(
    maybelog_levels: _np.ndarray,
    maybelog_changes: _np.ndarray,
    /,
    default_level: Number = DEFAULT_INIT_GUESS,
    default_change: Number = 0,
) -> tuple[_np.ndarray, _np.ndarray]:
    """
    """
    #[
    maybelog_levels[_np.isnan(maybelog_levels)] = default_level
    maybelog_changes[_np.isnan(maybelog_changes)] = default_change
    return maybelog_levels, maybelog_changes
    #]


def _prepare_time_shifts(
    equations: Iterable[_equations.Equation],
    /,
) -> tuple[_np.ndarray, int]:
    """
    Return [min_shift, ..., max_shift] and the index of the column with shift=0
    """
    #[
    # Add 1 to max shift to accommodate nonflat steady equators
    add_shift = 1
    #
    min_shift = _equations.get_min_shift_from_equations(equations, )
    max_shift = _equations.get_max_shift_from_equations(equations, ) + add_shift
    num_columns = -min_shift + NONFLAT_STEADY_SHIFT + max_shift
    shift_in_first_column = min_shift
    shift_vec = _np.array(range(min_shift, max_shift+1, ), dtype=float, ).reshape(1, -1, )
    return shift_vec, min_shift
    #]





"""
Interface to X13-ARIMA-TRAMO-SEATS
"""


#[
from __future__ import annotations

from typing import (Literal, TYPE_CHECKING, )
import os as _os
import numpy as _np
import subprocess as _sp
import platform as _pf
import tempfile as _tf
import copy as _co
import glob as _gl
import documark as _dm

from .. import executables as _executables
from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants
from ..dates import (Period, Span, Frequency, )

from . import _functionalize

if TYPE_CHECKING:
    from typing import (Any, )
    from types import (EllipsisType, )
    from ..series import (Series, )
#]


__all__ = ()


Mode = Literal["mult", "add", "pseudoadd", "logadd", ]


class Inlay:
    """
    """
    #[

    @_dm.reference(category="filtering", )
    def x13(
        self,
        *,
        span: Span | EllipsisType = ...,
        when_error: _wrongdoings.HOW = "warning",
        clean_up: bool = True,
        output: str = "seasonally_adjusted",
        #
        return_info: bool = False,
        unpack_singleton: bool = True,
        #
        specs_template: dict[str, Any] | None = None,
        mode: Literal["mult", "add", "pseudoadd", "logadd", ] | None = None,
        allow_missing: bool = False,
        add_to_specs: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        r"""
................................................................................

==X13-ARIMA-TRAMO-SEATS seasonal adjustment procedure==


### Function for creating new Series objects ###


```
new = irispie.x13(
    self,

    span=None,
    output="seasonally_adjusted",
    mode=None,
    when_error="warning",
    clean_up=True,

    specs_template=None,
    add_to_specs=None,
    allow_missing=False,
    mode=None,

    unpack_singleton=True,
    return_info=False,
)
```

```
new, info = irispie.x13(
    ...,
    return_info=True,
    ...,
)
```


### Class methods for changing existing Series objects in-place ###


```
self.x13(
    self,

    span=None,
    output="seasonally_adjusted",
    mode=None,
    when_error="warning",
    clean_up=True,

    specs_template=None,
    add_to_specs=None,
    allow_missing=False,
    mode=None,

    unpack_singleton=True,
    return_info=False,
)
```

```
info = self.x13(
    ...,
    return_info=True,
    ...,
)
```


### Input arguments ###


???+ input "self"
    A time `Series` object whose data will be run through the
    X13-ARIMA-TRAMO-SEATS procedure.

???+ input "span"
    A time span be specified as a `Span` object. If `span=None` or `span=...`,
    the time span goes from the first observed period to the last observed
    period in the input time series.

???+ input "output"
    The type of output to be returned by X13. The following options are
    available at the moment:

    | Output                  | X13 table | Description
    |-------------------------|-----------|-------------
    | `"seasonal"`            | `d10`     | Seasonal factors
    | `"seasonally_adjusted"` | `d11`     | Seasonally adjusted series
    | `"trend_cycle"`         | `d12`     | Trend-cycle component
    | `"irregular"`           | `d13`     | Irregular component
    | `"seasonal_and_td"`     | `d16`     | Combined seasonal and trading day factors
    | `"holiday_and_td"`      | `d18`     | Combined holiday and trading day factors

???+ input "specs_template"
    A dictionary with a specs template for the X13 run; if `None`, a default
    specs template is used (see below for the structure of the default template).

???+ input "mode"
    The mode to be used for the X13 run. The following options are available (see the
    [X13 documentation](https://www.census.gov/srd/www/x13as/)):

    | Mode          | Description
    |---------------|-------------
    | `None`        | Automatically selected
    | `"mult"`      | Multiplicative
    | `"add"`       | Additive
    | `"pseudoadd"` | Pseudo-additive
    | `"logadd"`    | Log-additive

    If `mode=None`, the mode is automatically selected based on the data. If the data is
    strictly positive or strictly negative, the multiplicative mode is used, otherwise
    the additive mode is used.

???+ input "allow_missing"
    If `True`, allow missing values in the input time series and automatically
    add an empty `automdl` spec if no ARIMA model is specified.

???+ input "add_to_specs"
    A dictionary with additional settings to be added to the `specs_template` (or
    the default templated).

???+ input "when_error"
    The action to be taken when an error occurs. The following options are
    available:

    | Action      | Description
    |-------------|-------------
    | `"warning"` | Issue a warning
    | `"error"`   | Raise an error

???+ input "unpack_singleton"
    If `True`, unpack `info` into a plain dictionary for models with a
    single variant.

???+ input "return_info"
    If `True`, return a dictionary with information about the X13 run as another
    output argument.


### Returns ###


???+ returns "self"
    The `Series` object with the output data.

???+ returns "new"
    A new `Series` object with the output data.

???+ returns "info"
    (Only returned if `return_info=True` which is not the default behavior)
    Dictionary with information about the X13 run; `info` contains the
    following items:

    | Key | Description
    |-----|-------------
    | `success` | True if the X13 run was successful
    | `specs_template` | The specs template used for the X13 run
    | `mode` | The mode used for the X13 run
    | `spc` | The spc file from the X13 run
    | `log` | The log file from the X13 run
    | `out` | The output file from the X13 run
    | `err` | The error file from the X13 run
    | `*` | Any other output file written by X13


### Details ###


???+ abstract "Default SPC template structure"

    The default specs template is a dictionary equivalent to the following SPC
    file:

    ```
    series{
        start=$(series_start)
        data=(
    $(series_data)
        )
        period=$(series_period)
        decimals=5
        precision=5
    }

    transform{
        function=$(transform_function)
    }

    x11{
        mode=$(x11_mode)
        save=$(x11_save)
    }

    ```

................................................................................
        """
        span = tuple(self._resolve_dates(span))
        base_start, base_end = span[0], span[-1]
        self.clip(base_start, base_end, )
        #
        mode, transform_function, flip_sign = _resolve_mode(self, mode, )
        basic_settings = _prepare_basic_settings(
            base_start, output,
            mode=mode,
            transform_function=transform_function,
        )
        specs_template = _get_specs_template(specs_template, )
        _update_specs_template_with_extra_settings(specs_template, add_to_specs, allow_missing, )
        specs_template_string = _print_specs_template(specs_template, )
        specs_template_string = _insert_basic_settings_into_specs_template_string(specs_template_string, basic_settings, )
        #
        new_data = []
        out_info = []
        for variant_data in self.data.T:
            info_v = {
                "specs_template": specs_template,
                "mode": mode,
                "transform_function": transform_function,
                "flip_sign": flip_sign,
            }
            if flip_sign:
                variant_data = -variant_data
            new_data_v = _x13_data(
                specs_template_string,
                base_start,
                variant_data,
                basic_settings["x11_save"],
                info=info_v,
                clean_up=clean_up,
            )
            if flip_sign:
                new_data_v = -new_data_v
            new_data.append(new_data_v.reshape(-1, 1, ))
            out_info.append(info_v)
            #
        if not all(i["success"] for i in out_info):
            message = "X13 failed to produce a result for at least one variant."
            _wrongdoings.raise_as(when_error, message, )
        #
        new_data = _np.hstack(new_data, )
        self._replace_start_and_values(base_start, new_data, )
        #
        if return_info:
            out_info = _has_variants.unpack_singleton(
                out_info, self.is_singleton,
                unpack_singleton=unpack_singleton,
            )
            return out_info
        else:
            return
    #]


for n in ("x13", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _x13_data(
    specs: str,
    base_start: Period,
    base_variant_data: _np.ndarray,
    x11_save: str,
    /,
    *,
    info: dict[str, Any],
    clean_up: bool = True,
) -> tuple[_np.ndarray, dict[str, Any]]:
    """
    """
    #[
    new_data = _np.full(base_variant_data.shape, _np.nan, dtype=float, )
    start_date, variant_data, output_slice = _remove_leading_trailing_nans(base_start, base_variant_data, )
    year, period = start_date.to_year_period()
    specs = specs.replace("$(series_start)", _X13_DATE_FORMAT_RESOLUTION[start_date.frequency].format(year=year, period=period, ), )
    specs = specs.replace("$(series_data)", _print_series_data(variant_data, ), )
    specs_file_name_without_ext = _write_specs_to_file(specs, )
    system_output = _execute(specs_file_name_without_ext, )
    raw_output_data, update_info = _collect_outputs(specs_file_name_without_ext, system_output, x11_save, )
    info.update({"specs": specs, })
    info.update(update_info, )
    if raw_output_data is not None:
        new_data[output_slice] = raw_output_data
    if clean_up:
        _clean_up(specs_file_name_without_ext, )
    return new_data
    #]


def _print_series_data(
    variant_data: _np.ndarray,
    /,
) -> str:
    """
    """
    _INDENT = " " * 8
    _NAN_REPLACEMENT = "-99999.0"
    data_str = "\n".join(
        _INDENT + (f"{v:g}" if not _np.isnan(v) else _NAN_REPLACEMENT)
        for v in variant_data
    )
    return data_str


def _execute(
    specs_file_name_without_ext: str,
    /,
) -> _sp.CompletedProcess:
    """
    """
    #[
    return _sp.run(
        [_X13_EXECUTABLE_PATH, specs_file_name_without_ext, ],
        stdout=_sp.PIPE,
        check=True,
    )
    #]


def _collect_outputs(
    specs_file_name_without_ext: str,
    output: _sp.CompletedProcess,
    out_table: str,
    /,
) -> tuple[_np.ndarray | None, dict[str, str]]:
    """
    """
    #[
    info = _read_out_files(specs_file_name_without_ext, )
    try:
        info[out_table] = str(info[out_table])
        out_data = _read_output_data(info[out_table], )
    except:
        out_data = None
        info[out_table] = None
    info["success"] = (
        output.returncode == 0
        and "ERROR" not in str(output.stdout)
        and out_data is not None
    )
    return out_data, info
    #]


def _read_output_data(
    out_table: str,
) -> tuple[tuple[float, ...], str]:
    """
    """
    return tuple(float(i) for i in out_table.split()[5::2])


def _read_out_files(
    specs_file_name_without_ext: str,
) ->  dict[str, str]:
    """
    """
    #[
    info = dict()
    for file_name in _gl.glob(specs_file_name_without_ext + ".*"):
        ext = file_name.split(".")[-1]
        encoding = None if ext != "err" else "latin-1"
        with open(file_name, "rt", encoding=encoding, ) as fid:
            info[ext] = fid.read()
    return info
    #]


def _write_specs_to_file(
    specs: str,
    /,
) -> str:
    """
    """
    with _tf.NamedTemporaryFile(dir=".", mode="wt", suffix=".spc", delete=False, ) as fid:
        specs_file_name = fid.name
        fid.write(specs)
    return specs_file_name.removesuffix(".spc")


def _remove_leading_trailing_nans(
    base_start: Period,
    variant_data: _np.ndarray,
    /,
) -> tuple[Period, _np.ndarray, slice]:
    """
    """
    #[
    start_date = base_start.copy()
    num_rows = variant_data.shape[0]
    num_leading = 0
    num_trailing = 0
    while _np.isnan(variant_data[0]):
        num_leading += 1
        variant_data = variant_data[1:]
    while _np.isnan(variant_data[-1]):
        num_trailling += 1
        variant_data = variant_data[:-1]
    return start_date + num_leading, variant_data, slice(num_leading, num_rows-num_trailing, )
    #]


_TRANFORM_FUNCTION_DISPATCH = {
    "mult": "log",
}


def _resolve_mode(
    self: Series,
    mode: Mode | None,
    transform_function: str | None = None,
) -> tuple[str, str, bool]:
    """
    """
    #[
    flip_sign = False
    if mode is None:
        is_sign_strict = _np.all(self.data > 0) or _np.all(self.data < 0)
        mode = "mult" if is_sign_strict else "add"
    if transform_function is None:
        transform_function = _TRANFORM_FUNCTION_DISPATCH.get(mode, "none", )
    flip_sign = _np.all(self.data < 0)
    return mode, transform_function, flip_sign
    #]


def _prepare_basic_settings(
    base_start: Period,
    output: str,
    mode: Mode,
    transform_function: str = "none",
) ->  dict[str, str]:
    """
    """
    #[
    return {
        "series_period": str(base_start.frequency.value),
        "x11_mode": str(mode),
        "x11_save": _X11_OUTPUT_RESOLUTION.get(output, output),
        "transform_function": str(transform_function),
    }
    #]


def _get_specs_template(
    specs_template: dict[str, Any] | None,
) -> dict[str, Any]:
    """Get specs template either from a custom dict or a default dict"""
    return _co.deepcopy(specs_template or _DEFAULT_TEMPLATE)


def _update_specs_template_with_extra_settings(
    specs_template: dict[str, Any] | None,
    add_to_specs: dict[str, Any] | None,
    allow_missing: bool,
) -> None:
    """Update specs template with extra custom settings"""
    #[
    if add_to_specs:
        for k, v in add_to_specs.items():
            if k in specs_template:
                specs_template[k].update(v, )
            else:
                specs_template[k] = v
    _add_arima_model_if_needed(specs_template, allow_missing, )
    #]


def _add_arima_model_if_needed(
    specs_template: dict[str, Any],
    allow_missing: bool,
) -> None:
    """Add arima model if not present and in-sample missing values are allowed"""
    #[
    if not allow_missing:
        return
    has_arima = (
        ("automdl" in specs_template and specs_template["automdl"] is not False)
        or ("arima" in specs_template and specs_template["arima"] is not False)
    )
    if not has_arima:
        specs_template["automdl"] = {}
    #]


def _print_specs_template(
    specs_template: dict[str, Any],
) -> str:
    """Print the specs template dict to a string"""
    #[
    specs_template_string = "\n"
    for k, v in specs_template.items():
        if v is False:
            continue
        add = "\n\n" + k.lower() + "{"
        if isinstance(v, dict) and v:
            for kk, vv in v.items():
                add += f"\n    {kk.lower()}={vv}"
        add += "\n}"
        specs_template_string += add
    specs_template_string += "\n"
    return specs_template_string
    #]


def _insert_basic_settings_into_specs_template_string(
    specs_template_string: str,
    settings: dict[str, str],
) -> str:
    """
    Replace placeholders with custom setting
    """
    #[
    for k, v in settings.items():
        specs_template_string = specs_template_string.replace(f"$({k})", v)
    return specs_template_string
    #]


def _clean_up(
    specs_file_name_without_ext: str,
    /,
) -> None:
    """
    """
    #[
    for f in _gl.glob(specs_file_name_without_ext + ".*"):
        _os.remove(f, )
    #]


_EXECUTABLES_PATH = _os.path.dirname(_executables.__file__)

_DEFAULT_SERIES_TEMPLATE = {
    "start": "$(series_start)",
    "data": "(\n$(series_data)\n    )",
    "period": "$(series_period)",
    "decimals": 5,
    "precision": 5,
}

_DEFAULT_TRANSFORM_TEMPLATE = {
    "function": "$(transform_function)",
}

_DEFAULT_X11_TEMPLATE = {
    "mode": "$(x11_mode)",
    "save": "$(x11_save)",
}

_DEFAULT_TEMPLATE = {
    "series": _DEFAULT_SERIES_TEMPLATE,
    "transform": _DEFAULT_TRANSFORM_TEMPLATE,
    "x11": _DEFAULT_X11_TEMPLATE,
}

_X11_OUTPUT_RESOLUTION = {
    "sf": "d10",
    "seasonal": "d10",
    "seasonal_factors": "d10",
    "sa": "d11",
    "seasonally_adjusted": "d11",
    "tc": "d12",
    "trend_cycle": "d12",
    "irr": "d13",
    "irregular": "d13",
    "seasonal_and_td": "d16",
    "holiday_and_td": "d18",
}

_X13_DATE_FORMAT_RESOLUTION = {
    Frequency.MONTHLY: "{year:04d}.{period:02d}",
    Frequency.QUARTERLY: "{year:04d}.{period:1d}",
}

_X13_EXECUTABLE_FILE = {
    "Windows": "x13aswin.exe",
    "Linux": "x13asunix",
    "Darwin": "x13asmac",
}[_pf.system()]


_X13_EXECUTABLE_PATH = _os.path.join(_EXECUTABLES_PATH, _X13_EXECUTABLE_FILE, )





"""
"""


#[
from __future__ import annotations

import re as _re

from . import common as _common
#]


_TYPE_PATTERN = _re.compile(r"(\w+)`(\w+)")
_LIST_PATTERN = _re.compile(_common.KEYWORD_PREFIX + r"list\(`(\w+)\)")


def resolve_lists(source: str, ) -> str:
    """
    """
    #[

    type_to_names = _create_type_to_names(source, )
    if type_to_names:
        source = _replace_lists(source, type_to_names, )
        source = _remove_types(source, )
    return source

    #]


def _create_type_to_names(source: str, ) -> dict[str, str]:
    """
    Collect all name`type and put then into a dict of {"type": {"name1", "name2", ...}}
    """
    #[

    type_to_names = dict()
    for match in _TYPE_PATTERN.finditer(source, ):
        name, _type = match.groups()
        if _type not in type_to_names:
            type_to_names[_type] = set()
        type_to_names[_type].add(name, )
    return type_to_names

    #]


def _replace_lists(
    source: str,
    type_to_names: dict[str, str],
) -> str:
    """
    Replace list(`type) with name1, name2, ...
    """
    #[

    def replace(match: _re.Match, ) -> str:
        _type, = match.groups()
        return ", ".join(type_to_names.get(_type, tuple(), ), )

    return _LIST_PATTERN.sub(replace, source, )
    #]


def _remove_types(source: str, ) -> str:
    """
    Replace name`type with name
    """
    #[

    def replace(match: _re.Match, ) -> str:
        name, *_ = match.groups()
        return name

    return _TYPE_PATTERN.sub(replace, source, )
    #]





"""
Mixin classes for displaying data management objects
"""


#[
from __future__ import annotations
#]


_REPEAT_SHORT_ROW = 2
_VERTICAL_ELLIPSIS = "⋮"


class Mixin:
    """
    """
    #[

    _short_rows_: int = 5

    def _get_header_view_(self, /, ):
        """
        """
        return (
            "", 
            self._get_first_line_view(),
            f"Description: \"{self.get_description()}\"",
            self._get_header_separator(),
        )

    def _get_footer_view_(self, /, ):
        return ("", )

    def _get_view(self, /, ):
        """
        """
        header_view = self._get_header_view_()
        content_view = self._get_content_view()
        footer_view = self._get_footer_view_()
        return header_view + content_view + footer_view

    def __invert__(self):
        """
        ~self for short view
        """
        header_view = self._get_header_view_()
        content_view = self._get_content_view()
        if len(content_view) > 2*self._short_rows_:
            content_view = (
                content_view[:self._short_rows_]
                + (self._get_short_row_(), )*_REPEAT_SHORT_ROW
                + content_view[-self._short_rows_:]
            )
        print("\n".join(header_view + content_view))

    def __repr__(self, /, ):
        """
        """
        return "\n".join(self._get_view())

    def __str__(self, /, ):
        """
        """
        return repr(self)

    #]





import os
import irispie as ir


_THIS_DIR = os.path.dirname(os.path.abspath(__file__, ), )
_DATA_PATH = os.path.join(_THIS_DIR, "x13_test.csv", )


db = ir.Databox.from_csv(
    _DATA_PATH,
    name_row_transform=lambda x: x.replace("DATE", "__quarterly__"),
    period_from_string=ir.Period.from_iso_string,
)


db.rename(
    source_names=("ND000334Q", ),
    target_names=("GDP_NSA", ),
)


def test_vanilla():
    sa, info = ir.x13(
        db["GDP_NSA"],
        return_info=True,
    )
    assert info["success"]
    assert sa.start == db["GDP_NSA"].start
    assert sa.end == db["GDP_NSA"].end


def test_seasonal_factors():
    sf, info = ir.x13(
        db["GDP_NSA"],
        output="seasonal",
        return_info=True,
    )
    assert info["success"]
    assert sf.start == db["GDP_NSA"].start
    assert sf.end == db["GDP_NSA"].end


def test_in_sample_missing():
    miss = db["GDP_NSA"].copy()
    miss[ir.start+10>>ir.start+15] = None
    sf, info = ir.x13(
        miss,
        allow_missing=True,
        return_info=True,
    )
    assert info["success"]
    assert sf.start == db["GDP_NSA"].start
    assert sf.end == db["GDP_NSA"].end


def test_with_forecast_specs():
    sa, info = ir.x13(
        db["GDP_NSA"],
        return_info=True,
        add_to_specs={
            "pickmdl": {},
            "outlier": {},
            "forecast": {"maxlead": 12, "maxback": 0, },
        },
    )
    assert info["success"]
    assert sa.start == db["GDP_NSA"].start
    assert sa.end == db["GDP_NSA"].end


if __name__ == "__main__":
    # test_vanilla()
    # test_seasonal_factors()
    # test_in_sample_missing()
    test_with_forecast_specs()




"""
"""


#[

from __future__ import annotations

from typing import Literal, Callable
from collections.abc import Iterable
from numbers import Real
import functools as _ft
import json as _js
import numpy as _np

from .. import equations as _equations
from .. import quantities as _quantities
from .. import sources as _sources
from .. import dates as _dates
from .. import has_variants as _has_variants
from ..has_variants import unpack_singleton_decorator as _unpack_singleton
from ..series import main as _series
from ..incidences import main as _incidence
from ..databoxes.main import Databox
from ..fords.solutions import EigenvalueKind, Solution
from ..fords import descriptors as _descriptors


from . import _flags

#]


"""
Quantities that are time series in model databoxes
"""
_TIME_SERIES_QUANTITY = _quantities.QuantityKind.ANY_VARIABLE | _quantities.QuantityKind.ANY_SHOCK


def _cast_as_output_type(func: Callable, ):
    """
    """
    #[
    @_ft.wraps(func)
    def _wrapper(*args, **kwargs):
        output = func(*args, **kwargs)
        output_type = kwargs.get("output_type", Databox, )
        return output_type(output, ) if output_type else output
    return _wrapper
    #]


def _unpack_singleton_in_dict(func: Callable, ):
    #[
    @_ft.wraps(func)
    def _wrapper(self, *args, **kwargs, ):
        unpack_singleton = kwargs.pop("unpack_singleton", True)
        output = func(self, *args, **kwargs)
        return self.unpack_singleton_in_dict(output, unpack_singleton=unpack_singleton, )
    return _wrapper
    #]


class Inlay:
    """
    Frontend getter methods for Simultaneous objects
    """
    #[

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_steady_levels(
        self,
        /,
        unpack_singleton: bool = True,
        kind: _quantities.QuantityKind | None = None,
        **kwargs,
    ) -> dict[str, list[Real] | Real]:
        """
················································································

==Get steady-state levels of variables==

```
steady_levels = self.get_steady_levels(
    *,
    round: int = None,
    unpack_singleton: bool = True,
    output_typ: type = Databox,
)
```

### Input arguments ###


???+ input "self"
    `Simultaneous` model object with a valid first-order solution.

???+ input "round"
    Number of decimal places to round the output values.

???+ input "unpack_singleton"
    If `True`, unpack singleton lists and return the single element. If
    `False`, return the list as is.

???+ input "output_type"
    Cast output as this type; the constructor for this type must accept a
    dictionary.


### Returns ###


???+ returns "steady_levels"
    Databox with the steady-state levels for each variable.

......................................................................
        """
        kind = kind if kind is not None else _resolve_steady_kind(**kwargs, )
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, kind, )
        return self._get_values_as_dict("levels", qids, **kwargs, )

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_steady_changes(
        self,
        /,
        unpack_singleton: bool = True,
        kind: _quantities.QuantityKind | None = None,
        **kwargs,
    ) -> dict[str, list[Real] | Real]:
        """
················································································

==Get steady-state changes of variables==

```
steady_changes = self.get_steady_changes(
    *,
    round: int = None,
    unpack_singleton: bool = True,
    output_typ: type = Databox,
)
```


### Input arguments ###


???+ input "self"
    `Simultaneous` model object with a valid first-order solution.

???+ input "round"
    Number of decimal places to round the output values.

???+ input "unpack_singleton"
    If `True`, unpack singleton lists and return the single element. If
    `False`, return the list as is.

???+ input "output_type"
    Cast output as this type; the constructor for this type must accept a
    dictionary.


### Returns ###


???+ returns "steady_changes"
    Databox with the steady-state changes for each variable, i.e. first
    differences or gross rates of change depending on the log status of
    each variable.

......................................................................
        """
        kind = kind if kind is not None else _resolve_steady_kind(**kwargs, )
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, kind, )
        return self._get_values_as_dict("changes", qids, **kwargs, )

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_steady(
        self,
        /,
        unpack_singleton: bool = True,
        kind: _quantities.QuantityKind | None = None,
        **kwargs,
    ) -> dict[str, list[tuple[Real, Real]] | tuple[Real, Real]]:
        """
        """
        kind = kind if kind is not None else _resolve_steady_kind(**kwargs, )
        qids = tuple(_quantities.generate_qids_by_kind(self._invariant.quantities, kind, ))
        levels = self._get_values_as_dict("levels", qids, **kwargs, )
        changes = self._get_values_as_dict("changes", qids, **kwargs, )
        return {
            k: list(zip(levels[k], changes[k]))
            for k in levels.keys()
        }

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_parameters(
        self,
        /,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> dict[str, Real]:
        """
        """
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.PARAMETER, )
        return self._get_values_as_dict("levels", qids, **kwargs, )

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_stds(
        self,
        /,
        kind: _quantities.QuantityKind | None = None,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> dict[str, Real]:
        """
        """
        std_qids = self._get_std_qids(kind=kind, )
        return self._get_values_as_dict("levels", std_qids, **kwargs, )

    def _get_std_qids(
        self,
        /,
        kind: _quantities.QuantityKind | None = None,
    ) -> tuple[int, ...]:
        """
        """
        default_kind = _quantities.QuantityKind.ANY_STD
        kind = default_kind & kind if kind is not None else default_kind
        return _quantities.generate_qids_by_kind(self._invariant.quantities, kind, )

    @_cast_as_output_type
    @_unpack_singleton_in_dict
    def get_parameters_stds(
        self,
        /,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> dict[str, Real]:
        """
        """
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.PARAMETER_OR_STD, )
        return self._get_values_as_dict("levels", qids, **kwargs, )

    @_cast_as_output_type
    def get_log_status(self, **kwargs, ) -> dict[str, bool]:
        return {
            qty.human: qty.logly
            for qty in self._invariant.quantities
            if qty.kind in _sources.LOGGABLE_VARIABLE
        }

    def get_initials(
        self,
        /,
    ) -> _incidence.Tokens:
        """
        Get required list of initial conditions
        """
        # Get the tokens of state vector with actual incidence in the
        # model, and lag them by one period to get initial conditions
        initial_tokens = [
            _incidence.Token(t.qid, t.shift-1)
            for t in self._invariant.dynamic_descriptor.solution_vectors.get_initials()
        ]
        qid_to_name = self.create_qid_to_name()
        qid_to_logly = {} # [^1]
        return _incidence.print_tokens(initial_tokens, qid_to_name, qid_to_logly, )
        # [^1] Do not wrap initial conditions in log(...)

    # REFACTOR:
    # Create a SteadyBoxable class
    def generate_steady_items(
        self,
        start_date: _dates.Dater,
        end_date: _dates.Dater,
        /,
        deviation: bool = False,
    ) -> Iterable[tuple[str, _series.Series | Real]]:
        """
        """
        num_columns = int(end_date - start_date + 1)
        shift_in_first_column = start_date.get_distance_from_origin()
        qid_to_name = self.create_qid_to_name()
        qid_to_description = self.create_qid_to_description()
        #
        array = self.create_some_array(
            deviation=deviation,
            num_columns=num_columns,
            shift_in_first_column=shift_in_first_column,
        )
        num_rows = array.shape[0]
        #
        qid_to_kind = self.create_qid_to_kind()
        remove_qids = tuple(
            qid
            for qid, kind in qid_to_kind.items()
            if kind not in _TIME_SERIES_QUANTITY
        )
        names = (
            qid_to_name[qid]
            for qid in range(num_rows, )
            if qid in remove_qids
        )
        descriptions = (
            qid_to_description[qid]
            for qid in range(num_rows, )
            if qid in remove_qids
        )
        for qid in qid_to_name.keys():
            if qid_to_kind[qid] in _TIME_SERIES_QUANTITY:
                yield qid_to_name[qid], _series.Series(
                    start_date=start_date,
                    values=array[qid, :].reshape(-1, 1),
                    description=qid_to_description[qid],
                )
            else:
                yield qid_to_name[qid], array[qid, 0]

    def get_solution_vectors(self, /, ) -> _descriptors.HumanSolutionVectors:
        """
        Get the solution vectors of the model
        """
        qid_to_name = self.create_qid_to_name()
        qid_to_logly = self.create_qid_to_logly()
        return _descriptors.HumanSolutionVectors(
            self.solution_vectors,
            qid_to_name,
            qid_to_logly,
        )

    def get_solution(
        self,
        /,
        unpack_singleton: bool = True,
    ) -> Solution | list[Solution]:
        """
        """
        solution_matrices = [ v.solution for v in self._variants ]
        return self.unpack_singleton(solution_matrices, unpack_singleton=unpack_singleton, )

    def get_singleton_solution(
        self,
        deviation: bool = False,
        *,
        vid: int = 0,
    ) -> Solution:
        """
        """
        solution = self._variants[vid].solution
        return (
            solution if not deviation
            else Solution.deviation_solution(solution, )
        )

    def get_dynamic_equations(
        self,
        /,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        return tuple(
            _equations.generate_equations_of_kind(self._invariant.dynamic_equations, kind)
            if kind else self._invariant.dynamic_equations
        )

    def get_human_equations(
        self,
        /,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[str]:
        def _concatenate(dynamic: str, steady: str) -> str:
            return f"{dynamic} !! {steady}" if steady != dynamic else dynamic
        zipper = zip(
            self.get_dynamic_equations(kind=kind, ),
            self.get_steady_equations(kind=kind, ),
        )
        return tuple(
            _concatenate(dynamic.human, steady.human, )
            for dynamic, steady in zipper
        )

    def get_std_qids_for_shock_qids(
        self: Self,
        shock_qids: Iterable[int],
        /,
    ) -> tuple[int]:
        """
        """
        return tuple(
            self._invariant.shock_qid_to_std_qid[qid]
            for qid in shock_qids
        )

    def get_steady_equations(
        self,
        /,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        return tuple(
            _equations.generate_equations_of_kind(self._invariant.steady_equations, kind)
            if kind else self._invariant.steady_equations
        )

    def get_quantities(
        self,
        /,
        *,
        kind: _quantities.QuantityKind | None = None,
    ) -> _quantities.Quantities:
        return tuple(
            _quantities.generate_quantities_of_kind(self._invariant.quantities, kind, )
            if kind else self._invariant.quantities
        )

    def get_names(
        self,
        /,
        *,
        kind: _quantities.QuantityKind | None = None,
    ) -> _quantities.Quantities:
        return tuple(q.human for q in self.get_quantities(kind=kind, ))

    def get_flags(
        self,
        /,
    ) -> _flags.Flags:
        return self._invariant._flags

    def get_context(
        self,
        /,
    ) -> dict[str, Callable]:
        return self._invariant._context

    def _get_values_as_dict(
        self,
        variant_attr: Literal["levels", "changes"],
        qids: Iterable[int],
        /,
        **kwargs,
    ) -> dict[str, Any]:
        """
        """
        custom_round = kwargs.pop("round", None, )
        def apply(value: Real) -> Real:
            return (
                round(value, custom_round)
                if custom_round is not None and value is not None
                else value
            )
        #
        qid_to_name = self.create_qid_to_name()
        out_dict = {}
        for qid in qids:
            out_dict[qid_to_name[qid]] = [
                apply(getattr(v, variant_attr, )[qid], )
                for v in self._variants
            ]
        return out_dict

    def get_quantity_descriptions(self, ) -> dict[str, str]:
        """
        """
        return _quantities.create_name_to_description(self._invariant.quantities, )

    def get_equation_descriptions(self, ) -> dict[str, str]:
        """
        """
        return _equations.create_human_to_description(self._invariant.dynamic_equations, )

    @_unpack_singleton
    def get_eigenvalues(
        self,
        /,
        *,
        kind: EigenvalueKind = EigenvalueKind.ALL,
        unpack_singleton: bool = True,
    ) -> tuple[Real, ...] | list[tuple[Real, ...]]:
        """
        Eigenvalues
        """
        return [
            tuple(e for e, s in zip(v.solution.eigenvalues, v.solution.eigenvalues_stability, ) if s in kind)
            for v in self._variants
        ]

    @_unpack_singleton
    def get_eigenvalues_stability(
        self,
        *,
        kind: EigenvalueKind = EigenvalueKind.ALL,
        unpack_singleton: bool = True,
    ):
        return [ 
            tuple(i for i in v.solution.eigenvalues_stability if i in kind)
            for v in self._variants
        ]

    def _get_variable_stability_for_variant(self, variant, /, ) -> dict[str, bool]:
        """
        """
        vec = self._invariant.dynamic_descriptor.solution_vectors
        qid_to_name = self.create_qid_to_name()
        return {
            qid_to_name[token.qid]: \
                variant.solution.transition_vector_stability[index] == EigenvalueKind.STABLE
            for index, token in enumerate(vec.transition_variables)
            if token.shift == 0
        } | {
            qid_to_name[token.qid]: \
                variant.solution.measurement_vector_stability[index] == EigenvalueKind.STABLE
            for index, token in enumerate(vec.measurement_variables)
            if token.shift == 0
        }

    @_unpack_singleton
    def get_variable_stability(
        self,
        *,
        unpack_singleton: bool = True,
    ) -> dict[str, bool] | list[dict[str, bool]]:
        """
        """
        return [
            self._get_variable_stability_for_variant(v, )
            for v in self._variants
        ]

    def generate_minus_control_quantities(self, /, ) -> tuple[int]:
        """
        """

    def map_name_to_minus_control_func(self, /, ) -> tuple[str]:
        """
        """
        minus_control_func = {
            True: lambda x, y: x / y,
            False: lambda x, y: x - y,
            None: lambda x, y: x - y,
        }
        kind = _quantities.ANY_VARIABLE | _quantities.ANY_SHOCK
        return {
            q.human: minus_control_func[q.logly]
            for q in self._invariant.quantities
            if q.kind in kind
        }

    #]


def _resolve_steady_kind(
    *,
    include_shocks: bool = False,
    **kwargs,
) -> _quantities.QuantityKind:
    """
    """
    #[
    return (
        _sources.LOGGABLE_VARIABLE if not include_shocks
        else _sources.LOGGABLE_VARIABLE_OR_ANY_SHOCK
    )
    #]




"""
Main time series class definition
"""


#[

from __future__ import annotations

from numbers import (Real, )
from collections.abc import (Iterable, Callable, )
from typing import (Self, Any, TypeAlias, NoReturn, )
from types import (EllipsisType, )
import numpy as _np
import scipy as _sp
import itertools as _it
import functools as _ft
import operator as _op
import copy as _cp
import textwrap as _tw
import documark as _dm

from ..conveniences import descriptions as _descriptions
from ..conveniences import copies as _copies
from ..conveniences import iterators as _iterators
from ..dates import (Period, Span, Frequency, EmptyRanger, )
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from .. import has_variants as _has_variants

from . import _temporal
from . import _filling
from . import _moving
from . import _conversions
from . import _hp
from . import _ar
from . import _x13

from . import _indexing
from . import _plotly
from . import _views
from . import _functionalize

from ._temporal import __all__ as _diffs_cums__all__
from ._temporal import *

from ._filling import __all__ as _fillings__all__
from ._filling import *

from ._hp import __all__ as _hp__all__
from ._hp import *

from ._ar import __all__ as _ar__all__
from ._ar import *

from ._x13 import __all__ as _x13__all__
from ._x13 import *

from ._moving import __all__ as _moving__all__
from ._moving import *

from ._conversions import __all__ as _conversions__all__
from ._conversions import *

#]


_ELEMENTWISE_FUNCTIONS = {
    "log": _np.log,
    "exp": _np.exp,
    "sqrt": _np.sqrt,
    "abs": _np.abs,
    "sign": _np.sign,
    "sin": _np.sin,
    "cos": _np.cos,
    "tan": _np.tan,
    "round": _np.round,
    "logistic": _sp.special.expit,
    "normal_cdf": _sp.stats.norm.cdf,
    "normal_pdf": _sp.stats.norm.pdf,
    "maximum": _np.maximum,
    "minimum": _np.minimum,
}

FUNCTION_ADAPTATIONS_TIMEWISE_NUMPY = (
    "sum",
    "mean",
)


# FIXME
FUNCTION_ADAPTATIONS_ELEMENTWISE = tuple(_ELEMENTWISE_FUNCTIONS.keys())

FUNCTION_ADAPTATIONS_NUMPY_APPLY = () # ("maximum", "minimum", "mean", "median", "abs", )
FUNCTION_ADAPTATIONS_BUILTINS = ("max", "min", )
FUNCTION_ADAPTATIONS = tuple(set(
    FUNCTION_ADAPTATIONS_ELEMENTWISE
))


__all__ = (
    ("Series", "shift", "series", )
    + _conversions__all__
    + _diffs_cums__all__
    + _fillings__all__
    + _hp__all__
    + _ar__all__
    + _moving__all__
    + _x13__all__
    + FUNCTION_ADAPTATIONS
)

Dates: TypeAlias = Period | Iterable[Period] | Span | EllipsisType | None
VariantsRequestType: TypeAlias = int | Iterable[int] | slice | None


def _get_date_positions(dates, base, num_periods, /, ):
    pos = tuple(_dates.date_index(dates, base))
    min_pos = _builtin_min((x for x in pos if x is not None), default=0)
    max_pos = _builtin_max((x for x in pos if x is not None), default=0)
    add_before = _builtin_max(-min_pos, 0)
    add_after = _builtin_max(max_pos - num_periods + 1, 0)
    pos_adjusted = [
        p + add_before if p is not None else None
        for p in pos
    ]
    return pos_adjusted, add_before, add_after


@_dm.reference(
    path=("data_management", "time_series.md", ),
    categories={
        "constructor": "Constructing new time series",
        "conversion": "Converting time series frequency",
        "manipulation": "Manipulating time series values",
        "homogenizing": "Homogenizing and extrapolating time series",
        "filtering": "Filtering time series",
        "moving": "Applying moving window functions",
        "temporal_change": "Calculating temporal change",
        "temporal_cumulation": "Calculating temporal cumulation",
    },
)
class Series(
    _indexing.Inlay,
    _conversions.Inlay,
    _temporal.Inlay,
    _filling.Inlay,
    _hp.Inlay,
    _ar.Inlay,
    _moving.Inlay,
    _x13.Inlay,
    _plotly.Inlay,
    _views.Inlay,
    #
    _descriptions.DescriptionMixin,
    _copies.Mixin,
):
    r"""
················································································

Time series
============

Time `Series` objects represent numerical time series, organized as rows of
observations stored in [`numpy`](https://numpy.org) arrays and time stamped
using [time `Periods`](periods.md). A `Series` object can hold multiple
variants of the data, stored as mutliple columns.

················································································
    """
    #[

    __slots__ = (
        "start",
        "data",
        "data_type",
        "metadata",
        "__description__",
    )

    _numeric_format: str = "15g"
    _short_str_format: str = ">15"
    _date_str_format: str = ">12"
    _missing = _np.nan
    _missing_str: str = "⋅"
    _test_missing_period = staticmethod(lambda x: _np.all(_np.isnan(x)))

    @_dm.reference(
        category="constructor",
        call_name="Series",
    )
    def __init__(
        self,
        /,
        *,
        num_variants: int = 1,
        data_type: type = _np.float64,
        description: str = "",
        start: Period | None = None,
        start_date: Period | None = None,
        periods: Iterable[Period] | None = None,
        dates: Iterable[Period] | None = None,
        frequency: Frequency | None = None,
        values: Any | None = None,
        func: Callable | None = None,
        populate: bool = True,
    ) -> None:
        """
················································································

==Create a new `Series` object==

```
self = Series(
    start=start,
    values=values,
)
```

```
self = Series(
    periods=periods,
    values=values,
)
```


### Input arguments ###


???+ input "start"

    The time [`Period`](periods.md) of the first value in the `values`.

???+ input "periods"

    An iterable of time [`Periods`](periods.md) that will be used to time stamp
    the `values`. The iterable can be e.g. a tuple, a list, a time
    [`Span`](spans.md), or a single time [`Period`](periods.md).

???+ input "values"

    Time series values, supplied either as a single values, a tuple of values,
    or a NumPy array.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        self.start = None
        self.data_type = data_type
        self.data = _np.full((0, num_variants), _np.nan, dtype=self.data_type, )
        self.metadata = {}
        self.__description__ = description
        #
        start = start_date if start is None else start
        periods = dates if periods is None else periods
        #
        if populate:
            test = (x is not None for x in (start, periods, values, func))
            populator = _SERIES_POPULATOR.get(tuple(test), _invalid_constructor)
            populator(
                self,
                start=start,
                periods=periods,
                frequency=frequency,
                values=values,
                func=func,
            )

    @classmethod
    def _guaranteed(
        klass,
        start: Period,
        values: _np.ndarray,
        description: str = "",
    ) -> Self:
        """
        """
        self = klass()
        self.start = start
        self.data = values
        self.__description__ = description
        self.trim()
        return self

    def reset(self, /, ) -> None:
        self.__init__(
            num_variants=self.num_variants,
            data_type=self.data_type,
        )

    def _create_periods_of_missing_values(self, num_rows=0) -> _np.ndarray:
        """
        """
        return _np.full(
            (num_rows, self.shape[1], ),
            _np.nan,
            dtype=self.data_type
        )

    @property
    @_dm.reference(category="property", )
    def shape(self, /, ) -> tuple[int, int]:
        """==Shape of time series data=="""
        return self.data.shape

    @property
    @_dm.reference(category="property", )
    def num_periods(self, /, ) -> int:
        """==Number of periods from the first to the last observation=="""
        return self.data.shape[0]

    @property
    @_dm.reference(category="property", )
    def num_variants(self, /, ) -> int:
        """==Number of variants (columns) within the `Series` object=="""
        return self.data.shape[1]

    @property
    def is_singleton(self, /, ) -> bool:
        """
        True for time series with only one variant
        """
        return _has_variants.is_singleton(self.num_variants, )

    @property
    @_dm.reference(category="property", )
    def span(self, ):
        """==Time span of the time series=="""
        return Span(self.start, self.end, ) if self.start else ()

    range = span

    @property
    @_dm.reference(category="property", )
    def from_until(self, ):
        """==Two-tuple with the start date and end date of the time series=="""
        return self.start, self.end

    @property
    @_dm.reference(category="property", )
    def periods(self, /, ) -> tuple[Period, ...]:
        """==N-tuple with the periods from the start period to the end period of the time series=="""
        return tuple(self.range, )

    dates = periods

    @property
    @_dm.reference(category="property", call_name="start", )
    def _start(self):
        """==Start date of the time series=="""
        raise NotImplementedError

    @property
    def start_period(self, /, ):
        return self.start

    start_date = start_period

    @property
    @_dm.reference(category="property", )
    def end(self):
        """==End period of the time series=="""
        return (
            self.start + self.data.shape[0] - 1
            if self.start else None
        )

    end_period = end
    end_date = end

    @property
    @_dm.reference(category="property", )
    def frequency(self, /, ):
        """==Date frequency of the time series=="""
        return (
            self.start.frequency
            if self.start is not None
            else Frequency.UNKNOWN
        )

    @property
    @_dm.reference(category="property", )
    def is_empty(self, ) -> bool:
        """==True if the time series is empty=="""
        return not self.data.size

    @property
    @_dm.reference(category="property", )
    def has_missing(self, /, ):
        """==True if the time series is non-empty and contains in-sample missing values=="""
        return (not self.is_empty) and _np.isnan(self.data).any()

    def any_missing(self, *args, ) -> bool:
        """
        """
        return self._func_missing(_np.any, *args, )

    def all_missing(self, *args, ) -> bool:
        """
        """
        return self._func_missing(_np.all, *args, )

    def count_missing(self, *args, ) -> int:
        """
        """
        return self._func_missing(_np.count_nonzero, *args, )

    def _func_missing(self, func, *args, ) -> bool:
        """
        """
        data = self.get_data(*args, )
        return func(_np.isnan(data))

    def set_data(
        self,
        dates: Dates,
        data: Any | Series,
        variants: VariantsRequestType = None,
        /,
    ) -> None:
        """
        """
        dates = self._resolve_dates(dates, )
        is_data_ndarray = isinstance(data, _np.ndarray)
        is_empty_data = (
            data is None
            or (is_data_ndarray and data.size == 0)
        )
        if not dates and is_empty_data:
            return
        data = (
            _reshape_numpy_array(data, )
            if is_data_ndarray else data
        )
        vids = self._resolve_variants(variants, )
        if not self.start:
            self.start = next(iter(dates), None, )
            self.data = self._create_periods_of_missing_values(num_rows=1, )
        pos, add_before, add_after = _get_date_positions(dates, self.start, self.shape[0], )
        self.data = self._create_expanded_data(add_before, add_after)
        if add_before:
            self.start -= add_before
        if hasattr(data, "get_data"):
            data = data.get_data(dates)
        #
        data_variants = _has_variants.iter_variants(data, )
        for c, d in zip(vids, data_variants, ):
            self.data[pos, c] = d
        self.trim()

    def _get_data_and_recreate(
        self,
        *args,
    ) -> _np.ndarray:
        """
        """
        dates, pos, variants, expanded_data = self._resolve_dates_and_positions(*args, )
        data = expanded_data[_np.ix_(pos, variants)]
        num_variants = data.shape[1]
        new = Series(num_variants=num_variants, data_type=self.data_type)
        new.set_data(dates, data)
        return new

    def _resolve_dates_and_positions(
        self,
        dates: Dates,
        variants: VariantsRequestType = None,
        /,
    ) -> tuple[Iterable[Period], Iterable[int], Iterable[int], _np.ndarray]:
        """
        """
        dates = [ t for t in self._resolve_dates(dates, ) ]
        variants = self._resolve_variants(variants)
        if not dates:
            dates = ()
            pos = ()
            data = self._create_periods_of_missing_values(num_rows=0, )[:, variants]
            return dates, pos, variants, data
        #
        base_date = self.start or _builtin_min(dates, )
        pos, add_before, add_after = _get_date_positions(dates, base_date, self.shape[0], )
        data = self._create_expanded_data(add_before, add_after, )
        if not isinstance(pos, Iterable):
            pos = (pos, )
        return dates, pos, variants, data

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> Self:
        """
        Alter (expand, shrink) the number of variants in this time series object
        """
        if new_num < self.num_variants:
            self.shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self.expand_num_variants(new_num, )

    def expand_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        """
        add = new_num - self.num_variants
        if add < 0:
            raise ValueError("Use shrink_num_variants to shrink the number of variants")
        data = _np.tile(self.data[:, (-1, )], (1, add, ))
        self.data = _np.hstack((self.data, data, ), )

    def shrink_num_variants(
        self,
        new_num: int,
        /,
    ) -> None:
        """
        """
        remove = self.num_variants - new_num
        if remove < 0:
            raise ValueError("Use expand_num_variants to expand the number of variants")
        self.data = self.data[:, :-remove]

    def get_values(
        self,
        *args,
        unpack_singleton: bool = True,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        data = self.get_data(*args, **kwargs, )
        values = [ tuple(data_column.tolist()) for data_column in data.T ]
        return _has_variants.unpack_singleton(
            values, self.is_singleton,
            unpack_singleton=unpack_singleton,
        )

    def get_data(self, *args, **kwargs, ) -> _np.ndarray:
        return self.get_data_and_periods(*args, **kwargs, )[0]

    def get_data_and_periods(
        self,
        dates: Iterable[Period] | None | EllipsisType = ...,
        *args,
    ) -> _np.ndarray:
        periods, pos, variants, expanded_values = self._resolve_dates_and_positions(dates, *args, )
        return expanded_values[_np.ix_(pos, variants)], periods

    def get_data_variant(
        self,
        dates: Dates,
        variant: Real | None = None,
        /,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant<self.data.shape[1] else 0
        return self.get_data(dates, variant, )

    def get_data_from_until(
        self,
        from_until,
        *args,
    ) -> _np.ndarray:
        """
        """
        _, pos, variants, expanded_data \
            = self._resolve_dates_and_positions(from_until, *args, )
        from_pos, to_pos = pos[0], pos[-1]+1
        return expanded_data[from_pos:to_pos, variants]

    def get_data_variant_from_until(
        self,
        from_until: Iterable[Period],
        variant: int | None = None,
        /,
    ) -> _np.ndarray:
        """
        """
        variant = variant if variant and variant < self.data.shape[1] else 0
        return self.get_data_from_until(from_until, variant, )

    def extract_variants(
        self,
        variants,
        /,
    ) -> None:
        if not isinstance(variants, Iterable):
            variants = (variants, )
        else:
            variants = tuple(c for c in variants)
        self.data = self.data[:, variants]

    def set_start(
        self,
        new_start: Period,
        /,
    ) -> Self:
        self.start = new_start
        return self

    def _resolve_dates(self, dates, ) -> tuple[Period]:
        """
        """
        if dates is None:
            # dates = []
            dates = ...
        if isinstance(dates, slice) and dates == slice(None, ):
            dates = ...
        if dates is ... and self.start is not None:
            dates = Span(None, None, )
        if dates is ... and self.start is None:
            dates = EmptyRanger()
        if hasattr(dates, "needs_resolve") and dates.needs_resolve:
            dates = dates.resolve(self, )
        return tuple(
            d.resolve(self) if hasattr(d, "needs_resolve") and d.needs_resolve else d
            for d in dates
        )

    def _resolve_variants(
        self,
        variants: VariantsRequestType,
        /,
    ) -> Iterable[int]:
        """
        Resolve variant request to an iterable of integers
        """
        if variants is None or variants is Ellipsis:
            variants = slice(None, )
        if isinstance(variants, slice):
            variants = range(*variants.indices(self.num_variants))
        if not isinstance(variants, Iterable):
            variants = (variants, )
        return variants

    def shift(
        self,
        by: int | str = -1,
        /,
    ) -> None:
        match by:
            case "yoy":
                self._shift_by_number(-self.frequency.value, )
            case "boy" | "soy":
                self._shift_to_soy()
            case "eopy":
                self._shift_to_eopy()
            case "tty":
                self._shift_to_tty()
            case _:
                self._shift_by_number(by, )

    def _shift_by_number(self, by: int, ) -> None:
        """
        Shift (lag, lead) start date by a number of periods
        """
        self.start = (
            self.start - by
            if self.start else self.start
        )

    def _shift_to_soy(self, ) -> None:
        new_data = self.get_data(
            t.create_soy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_to_tty(self, ) -> None:
        new_data = self.get_data(
            t.create_tty()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_to_eopy(self, ) -> Self:
        new_data = self.get_data(
            t.create_eopy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def hstack(self, *args, ):
        """
        """
        if not args:
            return self.copy()
        self_args = (self, *args, )
        if all(i.is_empty for i in self_args):
            num_variants = sum(i.num_variants for i in self_args)
            return Series(num_variants=num_variants, )
        encompassing_span, *from_until = _dates.get_encompassing_span(self, *args, )
        new_data = self.get_data_from_until(from_until, )
        add_data = (
            x.get_data_from_until(from_until, )
            if hasattr(x, "get_data")
            else _create_data_variant_from_number(x, encompassing_span, self.data_type)
            for x in args
        )
        new_data = _np.hstack((new_data, *add_data))
        new = Series(num_variants=new_data.shape[1], )
        new.set_data(encompassing_span, new_data, )
        return new

    @_dm.reference(category="homogenizing", )
    def clip(
        self,
        /,
        new_start: Period | None,
        new_end: Period | None,
    ) -> None:
        r"""
················································································

==Clip time series to a new start and end period==

    self.clip(new_start, new_end)


### Input arguments ###


???+ input "new_start"
    The new start period for the `self` time series; if `None`, the current
    start period is kept.

???+ input "new_end"
    The new end period for the `self` time series; if `None`, the current
    end period is kept.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        if new_start is None or new_start < self.start:
            new_start = self.start
        if new_end is None or new_end > self.end:
            new_end = self.end
        if new_start == self.start and new_end == self.end:
            return
        self.data = self.get_data_from_until((new_start, new_end, ), )
        self.start = new_start

    def empty(
        self,
        *,
        num_variants: int | None = None
    ) -> None:
        num_variants = num_variants if num_variants is not None else self.num_variants
        self.data = _np.empty((0, num_variants), dtype=self.data_type)

    def overlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        self.set_data(other.range, other.data, )
        self.trim()

    def overlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        _broadcast_variants_if_needed(self, other, )
        self._LAY_METHOD_RESOLUTION[method]["overlay"](self, other, )

    def underlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        """
        """
        new_self = other.copy()
        new_self.overlay(self, )
        self._shallow_copy_data(new_self, )

    def underlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        _broadcast_variants_if_needed(self, other, )
        self._LAY_METHOD_RESOLUTION[method]["underlay"](self, other, )

    _LAY_METHOD_RESOLUTION = {
        "by_range": {"overlay": overlay_by_range, "underlay": underlay_by_range},
    }

    def redate(
        self,
        new_date: Period,
        old_date: Period | None = None,
    ) -> None:
        """
        """
        if old_data is None:
            self.start = new_date
        else:
            self.start = new_date - (old_date - self.start)

    @_dm.reference(category="manipulation", )
    def replace_where(
        self,
        test: Callable,
        new_value: Real,
    ) -> None:
        """
················································································

==Replace time series values that pass a test==

```
self.replace_where(
    test,
    new_value,
)
```


### Input arguments ###


???+ input "self"
    Time series whose observations will be tested and those passing the test
    replaced.

???+ input "test"
    A function (or a Callable) that takes a numpy array and returns `True` or
    `False` for each individual value.

???+ input "new_value"
    The value to replace the observations that pass the test.


### Returns ###


???+ returns "None"
    This method modifies `self` in-place and does not return a value.


················································································
        """
        self.data[test(self.data)] = new_value
        self.trim()

    def _shallow_copy_data(
        self,
        other: Self,
        /,
    ) -> None:
        """
        """
        for n in ("start", "data", "data_type", ):
            setattr(self, n, getattr(other, n, ))

    def __and__(self, other):
        """
        Implement the & operator as hstack
        """
        return self.hstack(other, )

    __or__ = __and__

    def trim(self):
        if self.data.size == 0:
            self.reset()
            return self
        num_leading, num_trailing \
            = _get_num_leading_trailing_missing_rows(self.data, )
        if num_trailing == self.data.shape[0]:
            self.reset()
            return self
        if not num_leading and not num_trailing:
            return self
        slice_from = num_leading or None
        slice_to = -num_trailing if num_trailing else None
        self.data = self.data[slice_from:slice_to, ...]
        if slice_from:
            self.start += int(slice_from)
        return self

    def _create_expanded_data(self, add_before, add_after):
        return _np.pad(
            self.data, ((add_before, add_after), (0, 0)),
            mode="constant", constant_values=_np.nan,
        )

    def _check_data_shape(self, data, /, ):
        if data.shape[1] != self.data.shape[1]:
            raise Exception("Time series data being assigned must preserve the number of variants")

    def __bool__(self, /, ):
        """
        """
        return self.data.size > 0

    def __neg__(self):
        """
        -self
        """
        output = self.copy()
        output.data = -output.data
        return output

    def __pos__(self):
        """
        +self
        """
        return self.copy()

    def __add__(self, other, ) -> Self | _np.ndarray | Real:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.add, )

    def __radd__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__radd__(other))

    def __mul__(self, other) -> Self|_np.ndarray:
        """
        self + other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.mul, )

    def __rmul__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__rmul__(other))

    def __sub__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.sub, )

    def __rsub__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rsub__(other))

    def __pow__(self, other):
        """
        self ** other
        """
        # FIXME: 1 ** numpy.nan -> 1 !!!!!
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.pow, )

    def __rpow__(self, other):
        """
        other ** self
        """
        return self.apply(lambda data: data.__rpow__(other))

    def __truediv__(self, other):
        """
        self - other
        """
        # arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, _op.truediv, )

    def __rtruediv__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rtruediv__(other))

    def __floordiv__(self, other):
        """
        self // other
        """
        return self._binop(other, _op.floordiv, )

    def __rfloordiv__(self, other):
        """
        other // self
        """
        return self.apply(lambda data: data.__rfloordiv__(other))

    def __mod__(self, other):
        """
        self % other
        """
        return self._binop(other, _op.mod, )

    def __rmod__(self, other, /, ):
        """
        other % self
        """
        return self.apply(lambda data: data.__rmod__(other))

    for n in ["gt", "lt", "ge", "le", "eq", "ne", ]:
        exec(f"def __{n}__(self, other): return self._binop(other, _op.{n}, )", )

    def apply(self, func, /, *args, **kwargs, ):
        new_data = func(self.data, *args, **kwargs, )
        axis = kwargs.get("axis", None, )
        if new_data.shape == self.data.shape:
            new = self.copy()
            new._replace_data(new_data, )
            return new
        elif (axis is None or axis == 1) and new_data.shape == (self.data.shape[0], ):
            new = self.copy()
            new.data = new_data.reshape(self.data.shape[0], 1, )
            new.trim()
            return new
        elif axis is None and new_data.shape == ():
            return new_data
        elif axis == 0 and new_data.shape == (self.data.shape[1], ):
            return new_data
        else:
            raise _wrongdoings.IrisPieError(
                "Function applied on a time series resulted"
                " in a data array with an unexpected shape"
            )

    def _binop(self, other, func, /, new=None, ):
        if not isinstance(other, type(self)):
            return self.apply(lambda data: func(data, other))
        # FIXME: empty encompassing range
        _, *from_until = _dates.get_encompassing_span(self, other)
        self_data = self.get_data_from_until(from_until, )
        other_data = other.get_data_from_until(from_until, )
        new_data = func(self_data, other_data)
        new = Series(num_variants=new_data.shape[1], ) if new is None else new
        new._replace_start_and_values(from_until[0], new_data, )
        return new

    def _broadcast_variants(self, num_variants, /, ) -> None:
        """
        """
        if self.data.shape[1] == num_variants:
            return
        if self.data.shape[1] == 1:
            self.data = _np.repeat(self.data, num_variants, axis=1, )
            return
        raise _wrongdoings.IrisPieError("Cannot broadcast variants")

    def _replace_data(
        self,
        new_values,
        /,
    ) -> None:
        """
        """
        self.data = new_values
        self.trim()

    def _replace_start_and_values(
        self,
        new_start,
        new_values,
        /,
    ) -> None:
        """
        """
        self.start = new_start
        self._replace_data(new_values, )

    def iter_dates_values(self, /, unpack_singleton=True, ):
        """
        Default iterator is line by line, yielding a tuple of (date, values)
        """
        def _unpack_singleton_data_row(data_list: list[Real], /, ):
            return _has_variants.unpack_singleton(data_list, True, )
        def _keep_data_row(data_list: list[Real], /, ):
            return data_list
        data_row_func = (
            _unpack_singleton_data_row
            if self.is_singleton and unpack_singleton
            else _keep_data_row
        )
        for date, data_row in zip(self.range, self.data, ):
            yield date, data_row_func(data_row.tolist(), )

    def iter_variants(self, /, ) -> Iterator[Self]:
        """
        """
        for data in self.iter_data_variants_from_until(..., ):
            new = self.copy()
            new._replace_data(data, )
            yield new

    def iter_own_data_variants_from_until(self, from_until, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        data_from_until = self.data if from_until == ... else self.get_data_from_until(from_until, )
        return iter(data_from_until.T, )

    def iter_data_variants_from_until(self, from_until, ) -> Iterator[_np.ndarray]:
        """
        Iterates over the data variants from the given start date to the given end date
        """
        return _iterators.exhaust_then_last(self.iter_own_data_variants_from_until(from_until, ), )

    def logistic(self, /, ) -> Self:
        """
        """
        self.data = 1 / (1 + _np.exp(-self.data))

    for n in FUNCTION_ADAPTATIONS_ELEMENTWISE:
        exec(f"def {n}(self, *args, **kwargs, ): self.data = _ELEMENTWISE_FUNCTIONS['{n}'](self.data, *args, **kwargs, )")

    for n in FUNCTION_ADAPTATIONS_TIMEWISE_NUMPY:
        exec(f"def {n}(self, *args, **kwargs, ): self.data = _np.{n}(self.data, axis=1, *args, **kwargs, ).reshape(-1, 1)")

    #]


def _get_num_leading_trailing_missing_rows(data: _np.ndarray, /, ):
    """
    """
    #[
    # Boolean index of rows with at least one observation
    boolex_observations = ~_np.all(_np.isnan(data, ), axis=1, )
    if _np.all(boolex_observations, ):
        # All rows have at least one observation
        num_leading = 0
        num_trailing = 0
    elif not _np.any(boolex_observations, ):
        # No row has any observation
        num_leading = 0
        num_trailing = data.shape[0]
    else:
        # Some rows have observations
        num_leading = _np.argmax(boolex_observations)
        num_trailing = _np.argmax(boolex_observations[::-1])
    return num_leading, num_trailing
    #]


def hstack(first, *args) -> Self:
    return first.hstack(*args)


def _create_data_variant_from_number(
    number: Real,
    span: Span,
    data_type: type,
    /,
) -> _np.ndarray:
    return _np.full((len(span), 1), number, dtype=data_type)


class series:
    """
    """
    #[

    for n in FUNCTION_ADAPTATIONS_TIMEWISE_NUMPY:
        exec(_tw.dedent(f"""
            @staticmethod
            def {n}(self: Series, *args, **kwargs, ) -> None:
                new = self.copy()
                new.{n}(*args, **kwargs, )
                return new
        """))

    #]


for n in FUNCTION_ADAPTATIONS_ELEMENTWISE:
    exec(
        f"@_ft.singledispatch\n"
        f"def {n}(x, *args, **kwargs):\n"
        f"    return _ELEMENTWISE_FUNCTIONS['{n}'](x, *args, **kwargs)\n"
    )
    exec(
        f"@{n}.register(Series, )\n"
        f"def _(x, *args, **kwargs):\n"
        f"    new = x.copy()\n"
        f"    new.{n}(*args, **kwargs)\n"
        f"    return new\n"
    )


for n in FUNCTION_ADAPTATIONS_BUILTINS:
    exec(
        f"_builtin_{n} = {n}"
    )
#     exec(
#         f"@_ft.singledispatch"
#         f"\n"
#         f"def {n}(x, *args, **kwargs): "
#         f"return _builtin_{n}(x, *args, **kwargs)"
#     )
#     exec(
#         f"@{n}.register(Series, )"
#         f"\n"
#         f"def {n}(x, *args, **kwargs): "
#         f"return x._{n}_(*args, **kwargs)"
#     )


def _from_periods_and_values(
    self,
    periods: Iterable[Period] | str,
    values: _np.ndarray | Iterable,
    frequency: Frequency | None = None,
    **kwargs,
) -> None:
    """
    """
    #[
    # dates = _dates.ensure_period_tuple(dates, frequency=frequency, )
    self.set_data(periods, values, )
    #]


def _from_periods_and_func(
    self,
    periods: Iterable[Period] | str,
    func: Callable,
    frequency: Frequency | None = None,
    **kwargs,
) -> Self:
    """
    Create a new time series from dates and a function
    """
    #[
    # dates = _dates.ensure_period_tuple(dates, frequency=frequency, )
    data = [
        [func() for j in range(self.num_variants)]
        for i in range(len(periods))
    ]
    data = _np.array(data, dtype=self.data_type)
    self.set_data(periods, data)
    #]


def _from_start_and_values(
    self,
    start: Period,
    values: _np.ndarray | Iterable,
    frequency: Frequency | None = None,
    **kwargs,
) -> None:
    """
    """
    #[
    # start = _dates.ensure_period_tuple(start, frequency=frequency, )[0]
    self.start = start
    if isinstance(values, _np.ndarray):
        values = _reshape_numpy_array(values, )
    else:
        values = _has_variants.iter_variants(values, )
        values = _np.column_stack(tuple(
            v for v, _ in zip(values, range(self.num_variants), )
        ))
    values = values.astype(self.data_type, )
    self.data = values
    self.trim()
    #]


def _reshape_numpy_array(values: _np.ndarray, /, ) -> _np.ndarray:
    """
    """
    #[
    return (
        values.reshape(values.shape[0], -1)
        if values.ndim >= 2 else values.reshape(-1, 1)
    )
    #]


def _broadcast_variants_if_needed(
    self: Series,
    other: Series,
    /,
) -> tuple[Series, Series]:
    """
    """
    #[
    if self.num_variants == other.num_variants:
        return self, other
    #
    if self.num_variants == 1:
        return self._broadcast_variants(other.num_variants, ), other
    #
    if other.num_variants == 1:
        return self, other._broadcast_variants(self.num_variants, )
    #
    raise _wrongdoings.IrisPieError("Cannot broadcast time series variants")
    #]


def _invalid_constructor(
    self,
    **kwargs,
) -> NoReturn:
    raise _wrongdoings.IrisPieError("Invalid Series object constructor")


_SERIES_POPULATOR = {
    (False, False, False, False): lambda self, **kwargs: None,
    (True, False, True, False): _from_start_and_values,
    (False, True, True, False): _from_periods_and_values,
    (False, True, False, True): _from_periods_and_func,
}


def shift(
    self,
    by: int | str = -1,
) -> _series.Series:
    """
    """
    #[
    new = self.copy()
    new.shift(by)
    return new
    #]


for n in ("underlay", "overlay", "redate", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )




"""
Printers for iterative steady-state solvers
"""


#[

from __future__ import annotations

from .. import quantities as _quantities
from .. import iter_printers as _iter_printers

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable
    from ..equations import Equation

#]


class _IterPrinter(_iter_printers.IterPrinter, ):
    """
    """
    #[

    def __init__(
        self,
        equations: Iterable[_equations.Equation],
        qids: Iterable[str],
        qid_to_logly: dict[str, bool],
        qid_to_name: dict[str, str],
        **kwargs,
    ) -> None:
        """
        """
        equation_strings = self._get_equation_strings(equations, )
        quantity_strings = self._get_quantity_strings(qids, qid_to_logly, qid_to_name, )
        super().__init__(
            equation_strings=equation_strings,
            quantity_strings=quantity_strings,
            **kwargs,
        )

    def _get_equation_strings(
        self,
        equations: Iterable[Equation],
        /,
    ) -> tuple[str, ...]:
        return tuple(i.human for i in equations)

    def _get_quantity_strings(*args, **kwargs, ) -> tuple[str, ...]:
        raise NotImplementedError

    #]


class FlatIterPrinter(_IterPrinter, ):
    """
    """
    #[

    def _get_quantity_strings(
        self,
        qids: Iterable[str],
        qid_to_logly: dict[str, bool],
        qid_to_name: dict[str, str],
        /,
    ) -> tuple[str, ...]:
        return tuple(
            _quantities.wrap_logly(qid_to_name[i], qid_to_logly.get(i, False), )
            for i in qids
        )

    #]


class NonflatIterPrinter(_IterPrinter, ):
    """
    """
    #[

    def _get_quantity_strings(
        self,
        qids: Iterable[str],
        qid_to_logly: dict[str, bool],
        qid_to_name: dict[str, str],
        /,
    ) -> tuple[str, ...]:
        return tuple(
            _quantities.wrap_logly(qid_to_name[i], qid_to_logly.get(i, False), )
            for i in qids
        ) + tuple(
            "∆" + _quantities.wrap_logly(qid_to_name[i], qid_to_logly.get(i, False), )
            for i in qids
        )

    #]




"""
Frequency conversion of time series
"""


#[

from __future__ import annotations

from typing import (Self, Callable, )
from types import (EllipsisType, )
import numpy as _np
import functools as _ft
import operator as _op
import statistics as _st
import documark as _dm

from .. import dates as _dates
from . import arip as _arip
from . import _functionalize

#]


__all__ = (
    "convert_roc",
    "convert_pct",
)


_builtin_min = min
_builtin_max = max
_builtin_sum = sum


_DEFAULT_METHOD = "mean"


class Inlay:
    """
    """
    #[

    @_dm.reference(category="conversion", )
    def aggregate(
        self,
        target_freq: _dates.Frequency,
        /,
        method: Literal["mean", "sum", "first", "last", "min", "max"] | Callable | None = None,
        remove_missing: bool = False,
        select: list[int] | None = None,
    ) -> None:
        """
················································································

==Aggregate time series to a lower frequency==


### Function form for creating new time `Series` objects ###

    new = irispie.aggregate(
        self,
        target_freq,
        /,
        method="mean",
        remove_missing=False,
        select=None,
    )


### Class method form for creating new time `Series` objects ###

    self.aggregate(
        target_freq,
        /,
        method="mean",
        remove_missing=False,
        select=None,
    )


### Input arguments ###


???+ input "target_freq"
    The new frequency to which the original time series will be diaggregated.

???+ input "method"
    Aggregation method, i.e. a function applied to the high-frequency
    values within each low-frequency period:

    | Method    | Description
    |-----------|-------------
    | "mean"    | Arithmetic average of high-frequency values
    | "sum"     | Sum of high-frequency values
    | "prod"    | Product of high-frequency values
    | "first"   | Value in the first high-frequency period
    | "last"    | Value in the last high-frequency period
    | "min"     | Minimum of high-frequency values
    | "max"     | Maximum of high-frequency values

???+ input "remove_missing"
    Remove missing values from the high-frequency data before
    applying the aggregation `method`.

???+ input "select"
    Select only the high-frequency values at the specified indexes;
    `select=None` means all values are used.


### Returns ###

???+ returns "self"
    The original time `Series` object with the aggregated data.

???+ returns "new"
    A new time `Series` object with the aggregated data.

················································································
        """
        method = method or _DEFAULT_METHOD
        method_func = (
            _AGGREGATION_METHOD_RESOLUTION[method]
            if isinstance(method, str) else method
        )
        #
        if target_freq == self.frequency:
            return
        if target_freq > self.frequency or target_freq is _dates.Frequency.UNKNOWN or self.frequency is _dates.Frequency.UNKNOWN:
            raise ValueError(f"Cannot aggregate from {self.frequency} frequency to {target_freq} frequency")
        #
        new_dater_class = _dates.PERIOD_CLASS_FROM_FREQUENCY_RESOLUTION[target_freq]
        #
        aggregate_within_data_func = _ft.partial(
            _aggregate_within_data,
            select,
            remove_missing,
            method_func,
        )
        #
        if self.frequency.is_regular:
            aggregate_func = _aggregate_regular_to_regular
        elif self.frequency is _dates.Frequency.DAILY:
            aggregate_func = _aggregate_daily_to_regular
        #
        new_start_date, new_data = aggregate_func(self, new_dater_class, aggregate_within_data_func, )
        self._replace_start_and_values(new_start_date, new_data, )

    @_dm.reference(category="conversion", )
    def disaggregate(
        self,
        target_freq: _dates.Frequency,
        /,
        method: str = "flat",
        **kwargs,
    ) -> Self:
        r"""
················································································

==Disaggregate time series to a higher frequency==


### Function form for creating new time `Series` objects ###

    new = irispie.disaggregate(
        self,
        target_freq,
        /,
        method="flat",
    )


### Class method form for changing existing time `Series` objects in-place ###

    self.disaggregate(
        target_freq,
        /,
        method="flat",
        model=None,
    )


### Input arguments ###


???+ input "target_freq"
    The new frequency to which the original time series will be aggregated.

???+ input "method"
    Aggregation method, i.e. a function applied to the high-frequency
    values within each low-frequency period:

    | Method    | Description
    |-----------|-------------
    | "flat"    | Repeat the high-frequency values
    | "first"   | Place the low-frequency value in the first high-frequency period
    | "last"    | Place the low-frequency value in the last high-frequency period
    | "arip"    | Interpolate using a smooth autoregressive process


### Returns ###


???+ returns "new"
    A new time `Series` object with the disaggregated data.



### Details ###

???+ details "ARIP algorithm"

    The `method="arip" setting invokes an interpolation method that assumes the
    underlying high-frequency process to be an autoregression. The method can be
    described in its state-space recursive form, although the numerical
    implementation is stacked-time.

    The `"rate"` model:

    $$
    \begin{gathered}
    x_t = \rho \, x_{t-1} + \epsilon_t \\[10pt]
    y_t = Z \, x_t \\[10pt]
    \epsilon_t \sim N(0, \sigma_t^2)
    \end{gathered}
    $$

    The `"diff"` model:

    $$
    \begin{gathered}
    x_t = x_{t-1} + c + \epsilon_t \\[10pt]
    y_t = Z \, x_t \\[10pt]
    \epsilon_t \sim N(0, 1)
    \end{gathered}
    $$

    where

    * $x_t$ is the underlying high-frequency process;

    * $y_t$ is the observed low-frequency time series;

    * $Z$ is an aggregation vector depending on the `aggregation` specification,

    | Aggregation | $Z$ vector
    |-------------|-----------
    | "sum"       | $(1, 1, \ldots, 1)$
    | "mean"      | $\tfrac{1}{n}\,(1, 1, \ldots, 1)$
    | "first"     | $(1, 0, \ldots, 0)$
    | "last"      | $(0, 0, \ldots, 1)$

    * $\rho$ is a gross rate of change estimated as the average rate of change
    in the observed series, $y_t$, and converted to high frequency;

    * $c$ is a constant estimated as the average difference in the observed
    series, $y_t$, and converted to high frequency;

    * $\sigma_t$ is a time-varying standard deviation of the high-frequency process, set to $\sigma_0 = 1$, and $\sigma_t = \rho \, \sigma_{t-1}$.

················································································
        """
        method_func = _CHOOSE_DISAGGREGATION_METHOD[method]
        #
        if target_freq == self.frequency:
            return
        #
        if target_freq < self.frequency \
            or target_freq is _dates.Frequency.UNKNOWN \
            or self.frequency is _dates.Frequency.UNKNOWN:
            #
            raise _wrongdoings.IrisPieCritical(
                f"Cannot disaggregate from {self.frequency} frequency to {target_freq} frequency"
            )
        #
        new_dater_class = _dates.PERIOD_CLASS_FROM_FREQUENCY_RESOLUTION[target_freq]
        new_start_date, new_data, *_ = method_func(self, new_dater_class, **kwargs, )
        self._replace_start_and_values(new_start_date, new_data, )

    #]


for n in ("aggregate", "disaggregate", ):
    exec(_functionalize.FUNC_STRING.format(n=n, ), globals(), locals(), )
    __all__ += (n, )


def _aggregate_daily_to_regular(
    self,
    new_dater_class: type,
    aggregate_within_data_func: Callable,
) -> Self:
    """
    """
    #[
    start_date = self.start_date.create_soy()
    end_date = self.end_date.create_eoy()
    from_until = (start_date, end_date, )
    start_year = self.start_date.get_year()
    end_year = self.end_date.get_year()
    new_start_date = new_dater_class.from_year_period(start_year, 1)
    new_end_date = new_dater_class.from_year_period(end_year, "end")
    get_slice_func = lambda t: slice(
        t.to_daily(position="start", ) - start_date,
        t.to_daily(position="end", ) - start_date + 1,
    )
    new_data = tuple(
        tuple(
            aggregate_within_data_func(data_variant[get_slice_func(t)])
            for data_variant in self.iter_own_data_variants_from_until(from_until, )
        )
        for t in _dates.Ranger(new_start_date, new_end_date)
    )
    #
    new_data = _np.array(new_data, dtype=self.data_type, )
    return new_start_date, new_data
    #]


def _aggregate_regular_to_regular(
    self,
    new_dater_class: type,
    aggregate_within_data_func: Callable,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    start_year = self.start_date.get_year()
    start_date = self.start_date.create_soy()
    end_date = self.end_date.create_eoy()
    new_start_date = new_dater_class.from_year_period(start_year, 1)
    target_freq = new_dater_class.frequency
    self_data = self.get_data_from_until((start_date, end_date))
    factor = self.frequency.value // target_freq
    #
    # Loop over variants, create within-period data, and aggregate
    multi_output_data_transposed = []
    for data_variant in self_data.T:
        output_data_transposed = tuple(
            aggregate_within_data_func(within_data)
            for within_data in data_variant.reshape((-1, factor), )
        )
        multi_output_data_transposed.append(output_data_transposed, )
    #
    multi_output_data = _np.array(multi_output_data_transposed, dtype=self.data_type, ).T
    return new_start_date, multi_output_data
    #]


def _disaggregate_flat(
    self,
    high_dater_class: type,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    high_freq = high_dater_class.frequency
    high_start_date = high_dater_class.from_ymd(*self.start_date.to_ymd(position="start", ), )
    factor = high_freq.value // self.frequency.value
    high_data = _np.repeat(self.data, factor, axis=0, )
    return high_start_date, high_data, factor
    #]


def _disaggregate_first(
    self,
    high_dater_class: type,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    high_start_date, high_data, factor = _disaggregate_flat(self, high_dater_class, )
    for i in range(1, factor, ):
        high_data[i::factor, :] = _np.nan
    return high_start_date, high_data
    #]


def _disaggregate_last(
    self,
    high_dater_class: type,
    /,
) -> tuple[Dater, _np.ndarray]:
    """
    """
    #[
    high_start_date, high_data, factor = _disaggregate_flat(self, high_dater_class, )
    for i in range(0, factor-1, ):
        high_data[i::factor, :] = _np.nan
    return high_start_date, high_data, factor
    #]


def _aggregate_within_data(
    select: list[int] | None,
    remove_missing: bool,
    method_func: Callable,
    within_data: _np.ndarray,
    /,
) -> _np.ndarray:
    """
    """
    #[
    if select is not None:
        select = tuple(select)
        within_data = within_data[select]
    if remove_missing:
        within_data = within_data[~_np.isnan(within_data)]
    return method_func(within_data) if within_data.size > 0 else _np.nan
    #]


_AGGREGATION_METHOD_RESOLUTION = {
    "mean": _st.mean,
    "geometric_mean": _st.geometric_mean,
    "sum": _builtin_sum,
    "prod": _np.prod,
    "first": _op.itemgetter(0),
    "last": _op.itemgetter(-1),
    "min": _builtin_min,
    "max": _builtin_max,
}


_CHOOSE_DISAGGREGATION_METHOD = {
    "flat": _disaggregate_flat,
    "first": _disaggregate_first,
    "last": _disaggregate_last,
    "arip": _arip.disaggregate_arip,
}


def convert_roc(
    roc: Real,
    from_freq: _dates.Frequency,
    to_freq: _dates.Frequency,
    /,
) -> Real:
    """
    """
    return float(roc) ** (float(from_freq) / float(to_freq))


def convert_pct(
    pct: Real,
    from_freq: _dates.Frequency,
    to_freq: _dates.Frequency,
    /,
) -> Real:
    """
    """
    return 100*(convert_roc(1 + float(pct)/100, from_freq, to_freq) - 1)


def convert_diff(
    diff: Real,
    from_freq: _dates.Frequency,
    to_freq: _dates.Frequency,
    /,
) -> Real:
    """
    """
    return float(diff) * (float(from_freq) / float(to_freq))




"""
Model equations
"""


#[

from __future__ import annotations

from typing import Self, Callable
from collections.abc import Iterable, Sequence
import enum as _en
import re as _re
import numpy as _np
import itertools as _it
import operator as _op
from dataclasses import dataclass

from .incidences.main import Token
from .incidences import main as _incidences
from . import quantities as _quantities
from . import attributes as _attributes
from . import wrongdoings as _wrongdoings

#]


EVALUATOR_PREAMBLE = "lambda x, t: "
_EVALUATOR_FORMAT = EVALUATOR_PREAMBLE + "[{joined_xtrings}]"
_REPLACE_UKNOWN = "?"


class EquationKind(_en.Flag):
    """
    Classification of model equations
    """
    #[

    TRANSITION_EQUATION = _en.auto()
    MEASUREMENT_EQUATION = _en.auto()
    STEADY_AUTOVALUES = _en.auto()

    ENDOGENOUS_EQUATION = TRANSITION_EQUATION | MEASUREMENT_EQUATION

    def to_keyword(self, /, ) -> str:
        return "!" + self.name.lower() + "s"

    def to_portable(self, /, ) -> str:
        return self.to_keyword().replace("!", "")

    @property
    def human(self, /, ) -> str:
        return self.name.replace("_", " ").title()

    #]


__all__  = (
    "TRANSITION_EQUATION",
    "MEASUREMENT_EQUATION",
    "STEADY_AUTOVALUES",
    "ENDOGENOUS_EQUATION",
)
for n in __all__:
    exec(f"{n} = EquationKind.{n}")


@dataclass
class Equation:
    """
    """
    #[

    id: int | None = None
    human: str | None = None
    kind: EquationKind | None = None
    description: str | None = None
    xtring: str | None = None
    incidence: Iterable[Token] | None = None
    entry: int | None = None
    attributes: set[str] | None = None

    def finalize(self, name_to_id: dict[str, int], ) -> None:
        """
        Finalize the equation by creating an xtring from the human string, and and composing the incidence matrix
        Any undeclared names are assumed to be caught before this step
        """
        self.xtring, self.incidence, *_ = xtring_from_human(self.human, name_to_id, )

    def create_equator_func(self, /, *, context: dict[str, Callable]) -> Callable:
        context = (dict(context) if context else {}) | {"__builtins__": {}}
        equator_func_string = create_equator_func_string([self.xtring], context, )
        eval(equator_func_string, context, )

    def copy(self, /, ) -> Self:
        """
        Shallow copy of the equation
        """
        return type(self)(**self.__dict__, )

    def __hash__(self, /, ) -> int:
        return hash(self.__repr__)

    has_attributes = _attributes.has_attributes

    #]


def generate_all_tokens_from_equations(equations: Iterable[Equation], /, ) -> Iterable[Token]:
    return _it.chain.from_iterable(eqn.incidence for eqn in equations)


def finalize_equations(
    equations: Iterable[Equation],
    name_to_id: dict[str, int],
) -> None:
    """
    """
    #[
    for eqn in equations:
        eqn.finalize(name_to_id, )
    #]


def generate_names_from_human(human: str) -> Iterable[str]:
    """
    Generate all names from a single human string
    """
    return (f[0] for f in _quantities.QUANTITY_OCCURRENCE_PATTERN.findall(human))


def generate_all_names_from_equations(equations: Iterable[Equation]) -> list[str]:
    """
    Extract all names from a list of equations
    """
    return generate_names_from_human(" ".join(e.human for e in equations))


def create_name_to_qid_from_equations(equations: Iterable[Equation]) -> dict[str, int]:
    """
    """
    all_names = sorted(list(set(generate_all_names_from_equations(equations))))
    return { name: qid for qid, name in enumerate(all_names) }


def create_equator_func_string(xtrings: str) -> str:
    """
    """
    return _EVALUATOR_FORMAT.format(joined_xtrings=" , ".join(xtrings))


def create_eid_to_wrt_tokens(
    equations: Iterable[Equation],
    all_wrt_tokens: Iterable[Token],
) -> dict[int, Iterable[Token], ]:
    """
    """
    #[
    eid_to_wrt_tokens = {}
    for eqn in equations:
        eid_to_wrt_tokens[eqn.id] = _incidences.sort_tokens(
            wrt for wrt in all_wrt_tokens
            if wrt in eqn.incidence
        )
    return eid_to_wrt_tokens
    #]


def create_human_to_eid(
    equations: Iterable[Equation],
    /,
) -> dict[str, int]:
    return { eqn.human: eqn.id for eqn in equations }


def xtring_from_human(
    human: str,
    name_to_id: dict[str, int],
) -> tuple[str, set[Token]]:
    """
    Convert human string to xtring and retrieve incidence tokens
    """
    #[
    tokens_list: list[Token] = []

    def _replace_human_with_x(match: _re.Match) -> str:
        name = match.group(1)
        qid = name_to_id[name]
        shift = _resolve_shift_str(match.group(2))
        new_token = Token(qid, shift)
        tokens_list.append(new_token)
        return new_token.print_xtring()

    xtring = human
    xtring = xtring.replace(":=", "=")
    xtring = _quantities.QUANTITY_OCCURRENCE_PATTERN.sub(_replace_human_with_x, xtring, )
    xtring = _postprocess_xtring(xtring)
    return xtring, set(tokens_list), tokens_list
    #]


def _resolve_shift_str(shift_str: str) -> int:
    return int(shift_str.replace("[", "",).replace("]", "")) if shift_str is not None else 0


def _postprocess_xtring(equation: str) -> str:
    #[
    equation = equation.replace("^", "**")
    equation = equation.replace(" ", "")
    equation = equation.replace("===", "=")
    lhs_rhs = equation.split("=", maxsplit=1, )
    if len(lhs_rhs)==2:
        equation = "-(" + lhs_rhs[0] + ")+" + lhs_rhs[1]
    return equation
    #]


def generate_all_eids(equations: Iterable[Equation], ) -> Iterable[int]:
    return (eqn.id for eqn in equations)


def generate_eids_by_kind(
    equations: Iterable[Equation],
    kind: EquationKind,
) -> Iterable[int]:
    return (eqn.id for eqn in equations if eqn.kind in kind)


def generate_equations_of_kind(
    equations: Iterable[Equation],
    kind: EquationKind | None,
) -> Iterable[Equation]:
    return (eqn for eqn in equations if kind is None or eqn.kind in kind)


def count_equations_of_kind(
    equations: Iterable[Equation],
    kind: EquationKind,
) -> int:
    return sum(1 for eqn in generate_equations_of_kind(equations, kind, ))


def sort_equations(equations: Iterable[Equation], ) -> Iterable[Equation]:
    return sorted(equations, key=_op.attrgetter("id"))


def get_min_shift_from_equations(equations: Iterable[Equation], ) -> int:
    return _incidences.get_min_shift(generate_all_tokens_from_equations(equations))


def get_max_shift_from_equations(
    equations: Iterable[Equation],
    /,
) -> int:
    return _incidences.get_max_shift(generate_all_tokens_from_equations(equations))


def validate_selection_of_equations(
    allowed_equations: Iterable[Equation],
    custom_equations: Iterable[Equation] | None,
    /,
) -> tuple[Iterable[Equation], Iterable[Equation]]:
    """
    """
    #[
    invalid_equations = list(set(custom_equations) - set(allowed_equations)) if custom_equations is not None else []
    custom_equations = list(custom_equations) if custom_equations is not None else list(allowed_equations)
    #]
    return custom_equations, invalid_equations


def lookup_eids_by_human_starts(
    equations: Iterable[Equation],
    human_starts: Iterable[str],
    /,
) -> tuple[Iterable[Equation], list[str]]:
    """
    """
    #[
    human_starts = list(human_starts)
    eids = [
        next((eqn.id for eqn in equations if eqn.human.startswith(hs)), None)
        for hs in human_starts
    ]
    valid_eids = [ i for i in eids if i is not None ]
    invalid_human_starts = [ h for h, i in zip(human_starts, eids) if i is None ]
    #]
    return valid_eids, invalid_human_starts


def calculate_incidence_matrix(
    equations: Iterable[Equation],
    num_quantities: int,
    token_within_quantities: Callable,
    /,
    data_type: type = bool,
) -> _np.ndarray:
    """
    """
    #[
    equations = tuple(equations)
    num_equations = len(equations)
    incidence_matrix = _np.zeros((num_equations, num_quantities), dtype=data_type, )
    for row_index, eqn in enumerate(equations):
        #
        # For each incidenc token in this equation, the
        # `token_within_quantities` function returns either an integer meaning
        # the incidence token is to be included in the incidence matrix in the
        # corresponding column, or None meaning the incidence token is not to be
        # in the incidence matrix. The Nones need to be filtered out.
        column_indices = tuple(
            token_within_quantities(inc_tok, )
            for inc_tok in eqn.incidence
        )
        column_indices = tuple(i for i in column_indices if i is not None)
        incidence_matrix[row_index, column_indices] = True
        #
    return incidence_matrix
    #]


def reorder_by_kind(equations: Iterable[Equation], /) -> Iterable[Equation]:
    return list(sorted(equations, key=lambda x: (x.kind.value, x.entry)))


def stamp_id(equations: Iterable[Equation], /) -> None:
    """
    """
    for i, e in enumerate(equations, ):
        e.id = i


def create_human_to_description(equations: Iterable[Equation]) -> dict[str, str]:
    return { eqn.human: eqn.description for eqn in equations }




"""
Exporting data to CSV sheets
"""


#[

from __future__ import annotations

from typing import (TYPE_CHECKING, )
import pickle as _pk
import csv as _cs
import numpy as _np
import itertools as _it
import dataclasses as _dc
import functools as _ft
import documark as _dm

from ..dates import (Frequency, Period, EmptySpan, )
from .. import wrongdoings as _wrongdoings

if TYPE_CHECKING:
    from collections.abc import (Iterable, )
    from typing import (Any, Callable, )
    from ..databoxes.main import (Databox, )

#]


_DEFAULT_ROUND = 12
_DEFAULT_DATE_FORMATTER = str


@_dc.dataclass
class _ExportBlock:
    """
    """
    #[

    databox: Databox | None = None
    frequency: Frequency | None = None
    periods: tuple[Period] | None = None
    names: tuple[str] | None = None
    #
    # Options
    total_num_data_rows: int | None = None
    description_row: bool | None = None
    delimiter: str | None = None
    numeric_format: str | None = None
    nan_str: str | None = None
    round: int | None = None
    date_formatter: Callable | None = None

    def __iter__(self, ):
        """
        """
        def _round(x, /, ):
            if self.round is None:
                return x
            else:
                return _np.round(x, self.round)
        #
        descriptions = _get_descriptions_for_names(self.databox, self.names, )
        num_data_columns = _get_num_data_columns_for_names(self.databox, self.names, )
        data_array = _get_data_array_for_names(self.databox, self.names, self.periods, )
        #
        # Empty row
        empty_row = ("", ) + ("", )*sum(num_data_columns) + ("", )
        #
        # Name row
        name_row = tuple(_it.chain.from_iterable( 
            (n, ) + ("*", )*(num_data_columns[i] - 1)
            for i, n in enumerate(self.names, )
        ))
        yield (_get_frequency_mark(self.frequency), ) + name_row + ("", )
        #
        # Description row
        if self.description_row:
            description_row = tuple(_it.chain.from_iterable( 
                (n, ) + ("*", )*(num_data_columns[i] - 1)
                for i, n in enumerate(descriptions, )
            ))
            yield ("", ) + description_row + ("", )
        #
        # Dates and data, row by row
        date_formatter = self.date_formatter or _DEFAULT_DATE_FORMATTER
        for date, data_row in zip(self.periods, data_array, ):
            yield \
                (date_formatter(date, ), ) \
                + tuple(x if not _np.isnan(x) else self.nan_str for x in _round(data_row).tolist()) \
                + ("", )
        #
        # Empty rows afterwards
        for _ in range(self.total_num_data_rows - len(self.periods), ):
            yield empty_row

    #]


class Inlay:
    """
    Databox inlay for writing databox time series to CSV file
    """
    #[
    @_dm.reference(
        category="import_export",
    )
    def to_csv(
        self,
        file_name: str,
        *,
        span: Iterable[Period] | None = None,
        frequency_span: dict | None = None,
        names: Iterable[str] | None = None,
        description_row: bool = False,
        frequency: Frequency | None = None,
        numeric_format: str = "g",
        nan_str: str = "",
        delimiter: str = ",",
        round: int = _DEFAULT_ROUND,
        date_formatter: Callable | None = None,
        csv_writer_settings: dict | None = {},
        when_empty: Literal["error", "warning", "silent"] = "warning",
        return_info: bool = False,
    ) -> dict[str, Any]:
        r"""
················································································

==Write Databox time series to a CSV file==


    self.to_csv(
        file_name,
        *,
        frequency_span=None,
        names=None,
        description_row=False,
        frequency=None,
        numeric_format="g",
        nan_str="",
        delimiter=",",
        round=None,
        date_formatter=None,
        csv_writer_settings={},
        when_empty="warning",
    )


### Input arguments ###


???+ input "file_name"
    Name of the CSV file where the data will be written.

???+ input "frequency_span"
    Specifies the frequencies and their corresponding date ranges for exporting
    data. If `None`, exports data for all available frequencies and their full
    date ranges in the databox.

???+ input "names"
    A list of series names to export. If `None`, exports all series for the 
    specified frequencies.

???+ input "description_row"
    If `True`, include a row of series descriptions in the CSV.

???+ input "frequency"
    Frequency of the data to export.

???+ input "numeric_format"
    The numeric format for data values, e.g., 'g', 'f', etc.

???+ input "nan_str"
    String representation for NaN values in the output.

???+ input "delimiter"
    Character to separate columns in the CSV file.

???+ input "round"
    Number of decimal places to round numeric values.

???+ input "date_formatter"
    Function to format date values. If `None`, SDMX string formatter is used.

???+ input "csv_writer_settings"
    Additional settings for the CSV writer.

???+ input "when_empty"
    Behavior when no data is available for export. Can be "error", "warning", or
    "silent".


### Returns ###


???+ returns "info"
    A dictionary with details about the export:

    * `names_exported`: Names of the series exported to the CSV file.


················································································
        """
        databox = self.shallow(source_names=names, )
        frequency_span = _resolve_frequency_span(databox, frequency_span, span, )
        frequency_names = _resolve_frequency_names(databox, frequency_span, )
        _catch_empty(frequency_span, frequency_names, when_empty, file_name, )
        #
        total_num_data_rows = _get_total_num_data_rows(frequency_span, )
        export_block_constructor = _ft.partial(
            _ExportBlock,
            databox=databox,
            total_num_data_rows=total_num_data_rows,
            description_row=description_row,
            delimiter=delimiter,
            numeric_format=numeric_format,
            nan_str=nan_str,
            round=round,
            date_formatter=date_formatter,
        )
        export_blocks = (
            export_block_constructor(
                frequency=f,
                periods=frequency_span[f],
                names=frequency_names[f],
            )
            for f in frequency_span.keys()
            if frequency_names[f]
        )
        #
        csv_writer_settings = csv_writer_settings or {}
        with open(file_name, "w+") as fid:
            writer = _cs.writer(fid, delimiter=delimiter, lineterminator="\n", **csv_writer_settings, )
            for row in zip(*export_blocks, ):
                writer.writerow(_it.chain.from_iterable(row))
        #
        info = {
            "names_exported": tuple(_it.chain.from_iterable(frequency_names.values())),
        }
        #
        if return_info:
            return info
        else:
            return

    def to_sheet(self, *args, **kwargs, ):
        return self.to_csv(*args, **kwargs, )

    def to_pickle(
        self,
        file_name: str,
        /,
        **kwargs,
    ) -> None:
        """
        """
        with open(file_name, "wb+") as fid:
            _pk.dump(self, fid, **kwargs, )
    #]


def _get_frequency_mark(frequency, ):
    return "__" + frequency.name.lower() + "__"


def _get_names_to_export(databox, frequency, names, ):
    """
    """
    #[
    names_from_databox = self.get_series_names_by_frequency(frequency)
    names = names_from_databox if names is None else names
    return [n for n in names if n in names_from_databox]
    #]


_DEFAULT_FREQUENCY_SPAN = {
    Frequency.YEARLY: ...,
    Frequency.HALFYEARLY: ...,
    Frequency.QUARTERLY: ...,
    Frequency.MONTHLY: ...,
    Frequency.DAILY: ...,
    Frequency.INTEGER: ...,
    Frequency.UNKNOWN: ...,
}


def _resolve_frequency_span(
    databox: Databox,
    frequency_span: dict[Frequency | int: Iterable[Period]],
    span: Iterable[Period] | None = None,
    /,
) -> tuple[dict[Frequency: tuple[Period]], int]:
    """
    """
    #[
    # Argument span overrides frequency_span
    if span is None:
        frequency_span = (
            frequency_span
            if frequency_span is not None
            else _DEFAULT_FREQUENCY_SPAN
        )
    else:
        span = tuple(span)
        frequency = span[0].frequency
        frequency_span = {frequency: span}
    # Remove Nones, ensure Frequency objects
    frequency_span = {
        Frequency(k): v
        for k, v in frequency_span.items()
        if v is not None
    }
    #
    # Resolve date span for each ...
    frequency_span = {
        k: v if v is not ... else databox.get_span_by_frequency(k, )
        for k, v in frequency_span.items()
    }
    #
    # Expand date spans into tuples, remove empty spans
    frequency_span = {
        k: tuple(i for i in v)
        for k, v in frequency_span.items()
        if v is not EmptySpan() or k is Frequency.UNKNOWN
    }
    #
    return frequency_span
    #]


def _resolve_frequency_names(
    databox: Databox,
    frequency_span: Frequency | None,
    /,
) -> dict[Frequency: tuple[str, ...]]:
    """
    """
    #[
    return {
        k: tuple(databox.get_series_names_by_frequency(k, ))
        for k in frequency_span.keys()
    }


def _get_total_num_data_rows(
    frequency_span: dict[Frequency: tuple[Period]],
    /,
) -> int:
    """
    """
    #
    # Find maximum number of data rows
    return max((len(v) for v in frequency_span.values()), default=0, )


def _get_descriptions_for_names(
    self,
    names: Iterable[str],
    /,
) -> tuple[str, ...]:
    """
    """
    return tuple(self[n].get_description() for n in names)


def _get_num_data_columns_for_names(
    self,
    names: Iterable[str],
    /,
) -> tuple[int, ...]:
    """
    """
    return tuple(self[n].shape[1] for n in names)


def _get_data_array_for_names(
    self,
    names: Iterable[str],
    periods: Iterable[Period],
    /,
) -> _np.ndarray:
    """
    """
    empty_lead = _np.empty((len(periods), 0, ), dtype=_np.float64, )
    return _np.hstack([empty_lead] + [self[n].get_data(periods, ) for n in names], )


def _catch_empty(
    frequency_span: dict[Frequency: tuple[Period]],
    frequency_names: dict[Frequency: tuple[str, ...]],
    when_empty: Literal["error", "warning", "silent"],
    file_name: str,
    /,
) -> None:
    """
    """
    #[
    if not frequency_span or all(len(v) == 0 for v in frequency_names.values()):
        _wrongdoings.raise_as(when_empty, f"No data exported to {file_name}", )
    #]

